from typing import Literal

from pydantic import Field, ValidationInfo, model_validator
from typing_extensions import Self

from libresvip.model.base import BaseModel

from .enums import PocketSingerLyricsLanguage


class PocketSingerBgmMetadata(BaseModel):
    bar_beat_count: int = Field(alias="barBeatCount")
    beat_segment_count: int = Field(alias="beatSegmentCount")
    bgm_description: str = Field(alias="bgmDescription")
    bpm: float
    display_name: str = Field(alias="displayName")
    file_md5: str = Field(alias="fileMD5")
    file_name: str = Field(alias="fileName")
    file_type: str = Field(alias="fileType")
    first_beat_offset: float = Field(alias="firstBeatOffset")

    @model_validator(mode="after")
    def extract_audio(self, info: ValidationInfo) -> Self:
        if (
            info.context is not None
            and info.context["extract_audio"]
            and not hasattr(info.context["path"], "protocol")
        ):
            archive_audio_path = f"{self.file_name}.{self.file_type}"
            if not (
                audio_path := (
                    info.context["path"].parent / f"{self.display_name}.{self.file_type}"
                )
            ).exists():
                audio_path.write_bytes(info.context["archive_file"].read(archive_audio_path))
        return self


class PocketSingerMetadata(BaseModel):
    ace_file_name: str = Field(alias="aceFileName")
    background_image_name: str | None = Field(None, alias="backgroundImageName")
    bgm_info: PocketSingerBgmMetadata | None = Field(None, alias="bgmInfo")


class PocketSingerPitchBend(BaseModel):
    pitch: float
    time: float


class PocketSingerEnvolopeItem(BaseModel):
    envolope: float
    time: float


class PocketSingerSinWave(BaseModel):
    release_len: float
    freq: float
    attack_vol: float
    attack_len: float
    amp: float
    start_pos: float
    release_vol: float


class PocketSingerNoteBase(BaseModel):
    start_time: float
    end_time: float
    identifier: int | None = None
    energy_envolope: list[PocketSingerEnvolopeItem] | None = None
    air_envolope: list[PocketSingerEnvolopeItem] | None = None
    tension_envolope: list[PocketSingerEnvolopeItem] | None = None
    falsetto_envolope: list[PocketSingerEnvolopeItem] | None = None


class PocketSingerBrNote(PocketSingerNoteBase):
    type: Literal["br"] = "br"


class PocketSingerNote(PocketSingerNoteBase):
    pitch_bends: list[PocketSingerPitchBend] = Field(default_factory=list, alias="pitchBends")
    consonant_time_head: list[float] | None = None
    phone: list[str] = Field(default_factory=list)
    grapheme_count: int = 1
    grapheme_index: int = 0
    language: PocketSingerLyricsLanguage = PocketSingerLyricsLanguage.CHINESE
    grapheme: str = Field(validation_alias="word")
    pitch: int
    type: Literal["phone"] = "phone"
    br: bool = False
    sin_wave: PocketSingerSinWave | None = None
    user_pitch: list[PocketSingerPitchBend] | None = None
    consonant_time_tail: list[float] | None = None
    br_note: PocketSingerBrNote | None = Field(None, alias="brNote")
    user_phone: list[str] | None = Field(None, alias="userPhone")
    valid: bool | None = None
    key: str | None = None
    scale: list[int] | None = None
    config: str | None = None
    is_edit_phone: bool | None = Field(None, alias="isEditPhone")


class PocketSingerRoleInfo(BaseModel):
    name: str = ""
    role_id: int = 0


class PocketSingerTrack(BaseModel):
    sound_effect: int = 0
    pan: float = 0
    ai_svs_mode: bool = Field(True, alias="AI_SVS_mode")
    lyric: str = ""
    solo: bool = False
    notes: list[PocketSingerNote] = Field(default_factory=list)
    language: PocketSingerLyricsLanguage = PocketSingerLyricsLanguage.CHINESE
    mix_info: str | None = None
    br_notes: list[PocketSingerBrNote] = Field(default_factory=list)
    role_info: PocketSingerRoleInfo = Field(default_factory=PocketSingerRoleInfo)
    mute: bool = False
    front: bool = False
    singer_volume: float = 1


class PocketSingerBgmTrack(BaseModel):
    end_time: float
    file_md5: str
    file_name: str
    file_type: str
    position: float
    start_time: float


class PocketSingerBgmInfo(BaseModel):
    tracks: list[PocketSingerBgmTrack] = Field(default_factory=list)
    mute: bool = False
    bgm_volume: float = 1
    solo: bool = False


class PocketSingerDebugInfo(BaseModel):
    version: str = "1.6.2"
    record_type: str = Field("create", alias="recordType")
    os: str = "16"
    device: str = "iPad8,3"
    build_version: str | None = Field(None, alias="buildVersion")
    platform: str = "iOS"
    user_language: PocketSingerLyricsLanguage | None = Field(None, alias="userLanguage")


class PocketSingerSongInfo(BaseModel):
    start: int
    first_beat_offset: float
    scale: list[int]
    key: str
    segment_of_beat: int
    operate_scale: list[int] = Field(alias="operateScale")
    bpm: float
    duration: float
    beat_of_bar: int
    name: str
    origin_duration: float | None = None
    origin_start: int | None = None
    author: str | None = None
    tuner: str | None = None
    song_id: int | None = None
    user_id: int | None = None


class PocketSingerProject(BaseModel):
    notes: list[PocketSingerNote] | None = None
    tracks: list[PocketSingerTrack] = Field(default_factory=list)
    bpm: float | None = None
    bgm_info: PocketSingerBgmInfo = Field(default_factory=PocketSingerBgmInfo)
    debug_info: PocketSingerDebugInfo = Field(default_factory=PocketSingerDebugInfo)
    version: int = 3
    timestamp: int | None = None
    song_info: PocketSingerSongInfo

    @model_validator(mode="after")
    def migrate_notes(self) -> Self:
        if self.notes is not None:
            self.version = 1
            self.tracks.append(
                PocketSingerTrack(
                    notes=self.notes,
                )
            )
        return self



import pathlib

from PySide6.QtCore import QUrl
from PySide6.QtGui import QDesktopServices

from __feature__ import snake_case, true_property  # isort:skip # noqa: F401


def open_path(path: pathlib.Path) -> None:
    output_url = QUrl.from_local_file(path)
    QDesktopServices.open_url(output_url)


def open_url(url: str) -> None:
    QDesktopServices.open_url(QUrl(url))



import dataclasses

from libresvip.core.constants import KEY_IN_OCTAVE
from libresvip.core.lyric_phoneme.japanese import is_kana, is_romaji
from libresvip.core.time_sync import TimeSynchronizer
from libresvip.core.warning_types import show_warning
from libresvip.model.base import (
    InstrumentalTrack,
    Note,
    ParamCurve,
    Project,
    SongTempo,
    TimeSignature,
)
from libresvip.utils.translation import gettext_lazy as _

from .constants import DEFAULT_PHONEME, OCTAVE_OFFSET, TICK_RATE
from .model import (
    VoiSonaAudioEventItem,
    VoiSonaAudioTrackItem,
    VoiSonaBeatItem,
    VoiSonaNoteItem,
    VoiSonaParameterItem,
    VoiSonaParametersItem,
    VoiSonaPointData,
    VoiSonaProject,
    VoiSonaScoreItem,
    VoiSonaSingingTrackItem,
    VoiSonaSongItem,
    VoiSonaSoundItem,
    VoiSonaTempoItem,
    VoiSonaTimeItem,
    VoiSonaTrack,
)
from .options import OutputOptions
from .voisona_pitch import generate_for_voisona


@dataclasses.dataclass
class VoiSonaGenerator:
    options: OutputOptions
    first_bar_length: int = dataclasses.field(init=False)
    time_synchronizer: TimeSynchronizer = dataclasses.field(init=False)

    def generate_project(self, project: Project) -> VoiSonaProject:
        voisona_project = VoiSonaProject()
        self.time_synchronizer = TimeSynchronizer(project.song_tempo_list)
        self.first_bar_length = int(project.time_signature_list[0].bar_length())
        default_time_signatures = self.generate_time_signatures(project.time_signature_list)
        default_tempos = self.generate_tempos(project.song_tempo_list)
        voisona_project.tracks.append(VoiSonaTrack())
        for i, track in enumerate(project.track_list):
            if isinstance(track, InstrumentalTrack):
                audio_track = VoiSonaAudioTrackItem(
                    name=f"Audio{i}",
                    audio_event=[
                        VoiSonaAudioEventItem(
                            path=track.audio_file_path,
                            offset=self.time_synchronizer.get_actual_secs_from_ticks(track.offset),
                        )
                    ],
                )
                voisona_project.tracks[0].track.append(audio_track)
            else:
                singing_track = VoiSonaSingingTrackItem(
                    name=f"Singer{i}",
                )
                song_item = VoiSonaSongItem(
                    beat=[default_time_signatures],
                    tempo=[default_tempos],
                    score=[VoiSonaScoreItem(note=self.generate_notes(track.note_list))],
                )
                singing_track.plugin_data.state_information.song = [song_item]
                voisona_project.tracks[0].track.append(singing_track)
                if log_f0 := self.generate_pitch(
                    track.edited_params.pitch, project.song_tempo_list
                ):
                    singing_track.plugin_data.state_information.parameter = [
                        VoiSonaParametersItem(log_f0=[log_f0])
                    ]
        return voisona_project

    def generate_tempos(self, tempos: list[SongTempo]) -> VoiSonaTempoItem:
        return VoiSonaTempoItem(
            sound=[
                VoiSonaSoundItem(
                    clock=round(tempo.position * TICK_RATE) if i else 0,
                    tempo=tempo.bpm,
                )
                for i, tempo in enumerate(tempos)
            ]
        )

    def generate_time_signatures(self, time_signatures: list[TimeSignature]) -> VoiSonaBeatItem:
        beat = VoiSonaBeatItem(
            time=[
                VoiSonaTimeItem(
                    clock=0,
                    beats=time_signatures[0].numerator,
                    beat_type=time_signatures[0].denominator,
                )
            ]
        )
        tick = 0.0
        prev_time_signature = time_signatures[0]
        for time_signature in time_signatures[1:]:
            if time_signature.bar_index > prev_time_signature.bar_index:
                tick += (
                    time_signature.bar_index - prev_time_signature.bar_index
                ) * prev_time_signature.bar_length()
            beat.time.append(
                VoiSonaTimeItem(
                    clock=int(tick * TICK_RATE),
                    beats=time_signature.numerator,
                    beat_type=time_signature.denominator,
                )
            )
            prev_time_signature = time_signature
        return beat

    def generate_notes(self, notes: list[Note]) -> list[VoiSonaNoteItem]:
        voisona_notes = []
        for note in notes:
            lyric = note.lyric
            phoneme = ""
            if note.pronunciation:
                phoneme = note.pronunciation
            elif not is_kana(lyric) and not is_romaji(lyric):
                phoneme = DEFAULT_PHONEME
                msg_prefix = _("Unsupported lyric: ")
                show_warning(f"{msg_prefix} {lyric}")
            voisona_notes.append(
                VoiSonaNoteItem(
                    clock=int(note.start_pos * TICK_RATE),
                    duration=int(note.length * TICK_RATE),
                    lyric=note.lyric,
                    pitch_octave=note.key_number // KEY_IN_OCTAVE + OCTAVE_OFFSET,
                    pitch_step=note.key_number % KEY_IN_OCTAVE,
                    syllabic=0,
                    phoneme=phoneme,
                )
            )
        return voisona_notes

    def generate_pitch(
        self, pitch: ParamCurve, tempo_list: list[SongTempo]
    ) -> VoiSonaParameterItem | None:
        if (data := generate_for_voisona(pitch, tempo_list, self.first_bar_length)) is not None:
            return VoiSonaParameterItem(
                length=data.length,
                data=[
                    VoiSonaPointData(
                        index=each.idx,
                        repeat=each.repeat,
                        value=each.value,
                    )
                    for each in data.events
                ],
            )



import dataclasses
import operator
from typing import cast

from libresvip.core.lyric_phoneme.chinese import get_pinyin_series
from libresvip.core.lyric_phoneme.chinese.vocaloid_xsampa import pinyin2xsampa
from libresvip.core.lyric_phoneme.japanese import to_romaji
from libresvip.core.lyric_phoneme.japanese.vocaloid_xsampa import (
    legato_chars,
    romaji2xsampa,
)
from libresvip.core.lyric_phoneme.korean.vocaloid_xsampa import hangul2xsampa
from libresvip.core.tick_counter import shift_beat_list, skip_tempo_list
from libresvip.core.time_sync import TimeSynchronizer
from libresvip.core.warning_types import show_warning
from libresvip.model.base import (
    InstrumentalTrack,
    Note,
    ParamCurve,
    Project,
    SingingTrack,
    SongTempo,
    TimeSignature,
)
from libresvip.utils.audio import audio_track_info
from libresvip.utils.translation import gettext_lazy as _

from .constants import (
    BPM_RATE,
    DEFAULT_CHINESE_PHONEME,
    DEFAULT_JAPANESE_PHONEME,
)
from .model import (
    VocaloidStyleTypes,
    Vsq3,
    Vsq3MCtrl,
    Vsq3MonoTrack,
    Vsq3MonoUnit,
    Vsq3MusicalPart,
    Vsq3Note,
    Vsq3ParameterNames,
    Vsq3Singer,
    Vsq3StereoTrack,
    Vsq3StereoUnit,
    Vsq3Tempo,
    Vsq3TimeSig,
    Vsq3TypeParamAttr,
    Vsq3TypePhonemes,
    Vsq3VsTrack,
    Vsq3VsUnit,
    Vsq3VVoice,
    Vsq3WavPart,
)
from .models.enums import VocaloidLanguage
from .options import OutputOptions
from .vocaloid_pitch import generate_for_vocaloid


@dataclasses.dataclass
class Vsq3Generator:
    options: OutputOptions
    first_bar_length: int = dataclasses.field(init=False)
    style_params: dict[str, int | list[int]] = dataclasses.field(init=False)
    time_synchronizer: TimeSynchronizer = dataclasses.field(init=False)

    def generate_project(self, project: Project) -> Vsq3:
        self.style_params = VocaloidStyleTypes().model_dump(by_alias=True)
        self.time_synchronizer = TimeSynchronizer(project.song_tempo_list)
        self.first_bar_length = round(project.time_signature_list[0].bar_length())
        vsqx = Vsq3()
        mixer = vsqx.mixer
        master_track = vsqx.master_track
        tick_prefix = int(project.time_signature_list[0].bar_length())
        master_track.time_sig = self.generate_time_signatures(
            project.time_signature_list, master_track.pre_measure
        )
        master_track.tempo = self.generate_tempos(project.song_tempo_list, tick_prefix)
        vsqx.v_voice_table.v_voice.append(
            Vsq3VVoice(
                comp_id=self.options.default_comp_id,
                v_voice_name=self.options.default_singer_name,
            )
        )
        vsqx.vs_track, mixer.vs_unit = self.generate_singing_tracks(
            [track for track in project.track_list if isinstance(track, SingingTrack)],
            tick_prefix,
        )
        if first_instrumental_track := next(
            (track for track in project.track_list if isinstance(track, InstrumentalTrack)),
            None,
        ):
            self.generate_instrumental_track(first_instrumental_track, vsqx, tick_prefix)
        return vsqx

    def generate_instrumental_track(
        self, track: InstrumentalTrack, vsqx: Vsq3, tick_prefix: int
    ) -> None:
        if (
            (track_info := audio_track_info(track.audio_file_path, only_wav=True)) is not None
            and track_info.sampling_rate == 44100
            and track_info.bit_depth == 16
        ):
            wav_part = Vsq3WavPart(
                part_name=track.title,
                file_path=track.audio_file_path,
                pos_tick=tick_prefix + track.offset,
                play_time=round(
                    self.time_synchronizer.get_actual_ticks_from_secs_offset(
                        track.offset, track_info.duration / 1000
                    )
                )
                - track.offset,
                sample_rate=track_info.sampling_rate,
                sample_reso=track_info.bit_depth,
                channels=track_info.channel_s,
            )
            if track_info.channel_s == 1:
                vsqx.mixer.mono_unit = Vsq3MonoUnit(
                    mute=int(track.mute),
                    solo=int(track.solo),
                )
                vsqx.mono_track = Vsq3MonoTrack(wav_part=[wav_part])
            elif track_info.channel_s == 2:
                vsqx.mixer.stereo_unit = Vsq3StereoUnit(
                    mute=int(track.mute),
                    solo=int(track.solo),
                )
                vsqx.stereo_track = Vsq3StereoTrack(wav_part=[wav_part])

    def generate_singing_tracks(
        self, track_list: list[SingingTrack], tick_prefix: int
    ) -> tuple[list[Vsq3VsTrack], list[Vsq3VsUnit]]:
        vs_track_list = []
        vs_unit_list = []
        if not len(track_list):
            track_list.append(SingingTrack(title="Track"))
        elif self.options.default_lang_id not in [
            VocaloidLanguage.SIMPLIFIED_CHINESE,
            VocaloidLanguage.JAPANESE,
            VocaloidLanguage.KOREAN,
        ]:
            show_warning(
                _(
                    'Phonemes of all notes were set to "la". Please use "Lyrics" -> "Convert Phonemes" in the menu of VOCALOID3 to reset them.'
                )
            )
        for track_index, track in enumerate(track_list):
            vsqx_track = Vsq3VsTrack(
                vs_track_no=track_index,
                track_name=track.title,
            )
            if track.note_list:
                musical_part = Vsq3MusicalPart(
                    pos_tick=tick_prefix,
                    play_time=track.note_list[-1].end_pos if track.note_list else 0,
                    note=self.generate_notes(track.note_list),
                    singer=[Vsq3Singer(v_bs=self.options.default_lang_id)],
                )
                musical_part.part_style.attr.extend(
                    Vsq3TypeParamAttr(
                        type_param_attr_id=param_name,
                        value=cast("int", param_value),
                    )
                    for param_name, param_value in self.style_params.items()
                    if param_value is not None
                )
                if pitch := self.generate_pitch(track.edited_params.pitch, track.note_list):
                    musical_part.m_ctrl = pitch
                vsqx_track.musical_part = [musical_part]
            vsqx_unit = Vsq3VsUnit(
                vs_track_no=track_index,
                mute=int(track.mute),
                solo=int(track.solo),
            )
            vs_track_list.append(vsqx_track)
            vs_unit_list.append(vsqx_unit)
        return vs_track_list, vs_unit_list

    def generate_tempos(self, song_tempos: list[SongTempo], tick_prefix: int) -> list[Vsq3Tempo]:
        song_tempos = skip_tempo_list(song_tempos, tick_prefix)
        return [
            Vsq3Tempo(
                pos_tick=song_tempo.position,
                bpm=int(song_tempo.bpm * BPM_RATE),
            )
            for song_tempo in song_tempos
        ]

    def generate_time_signatures(
        self, time_signatures: list[TimeSignature], measure_prefix: int
    ) -> list[Vsq3TimeSig]:
        time_signatures = shift_beat_list(time_signatures, measure_prefix)
        return [
            Vsq3TimeSig(
                pos_mes=time_signature.bar_index,
                nume=time_signature.numerator,
                denomi=time_signature.denominator,
            )
            for time_signature in time_signatures
        ]

    def generate_notes(self, notes: list[Note]) -> list[Vsq3Note]:
        note_list = []
        for note in notes:
            vsqx_note = Vsq3Note(
                pos_tick=note.start_pos,
                dur_tick=note.length,
                note_num=note.key_number,
                lyric=note.lyric,
            )
            vsqx_note.note_style.attr.extend(
                Vsq3TypeParamAttr(
                    type_param_attr_id=param_name,
                    value=cast("int", param_value),
                )
                for param_name, param_value in self.style_params.items()
                if param_value is not None
            )
            if note.lyric in legato_chars:
                vsqx_note.phnms = Vsq3TypePhonemes(value="-")
            elif self.options.default_lang_id == VocaloidLanguage.SIMPLIFIED_CHINESE:
                vsqx_note.lyric = " ".join(
                    get_pinyin_series([note.lyric], filter_non_chinese=False)
                )
                vsqx_note.phnms = Vsq3TypePhonemes(
                    value=pinyin2xsampa.get(vsqx_note.lyric, DEFAULT_CHINESE_PHONEME),
                )
            elif self.options.default_lang_id == VocaloidLanguage.JAPANESE:
                vsqx_note.phnms = Vsq3TypePhonemes(
                    value=romaji2xsampa.get(
                        to_romaji(vsqx_note.lyric),
                        DEFAULT_JAPANESE_PHONEME,
                    ),
                )
            elif self.options.default_lang_id == VocaloidLanguage.KOREAN:
                vsqx_note.phnms = Vsq3TypePhonemes(
                    value=hangul2xsampa(vsqx_note.lyric),
                )
            else:
                vsqx_note.phnms = Vsq3TypePhonemes(value=DEFAULT_CHINESE_PHONEME)
            note_list.append(vsqx_note)
        return note_list

    def generate_pitch(self, pitch: ParamCurve, notes: list[Note]) -> list[Vsq3MCtrl]:
        music_controls: list[Vsq3MCtrl] = []
        if pitch_raw_data := generate_for_vocaloid(
            pitch, notes, self.first_bar_length, self.time_synchronizer
        ):
            music_controls.extend(
                Vsq3MCtrl(
                    pos_tick=pbs_event.pos,
                    attr=Vsq3TypeParamAttr(
                        type_param_attr_id=Vsq3ParameterNames.PBS.value,
                        value=pbs_event.value,
                    ),
                )
                for pbs_event in pitch_raw_data.pbs
            )
            music_controls.extend(
                Vsq3MCtrl(
                    pos_tick=pit_event.pos,
                    attr=Vsq3TypeParamAttr(
                        type_param_attr_id=Vsq3ParameterNames.PIT.value,
                        value=pit_event.value,
                    ),
                )
                for pit_event in pitch_raw_data.pit
            )
            music_controls.sort(key=operator.attrgetter("pos_tick"))
        return music_controls



import dataclasses
import functools
import operator
from collections.abc import Iterator

from sortedcontainers import SortedKeyList

from libresvip.core.constants import DEFAULT_BPM
from libresvip.core.exceptions import NoTrackError
from libresvip.core.time_sync import TimeSynchronizer
from libresvip.model.base import (
    Note,
    Project,
    SingingTrack,
    SongTempo,
    TimeSignature,
)
from libresvip.utils.translation import gettext_lazy as _

from .model import UTAUNote, UTAUProject, UTAUTimeSignature
from .options import InputOptions
from .pitch_mode1 import (
    UtauMode1NotePitchData,
    UtauMode1TrackPitchData,
    pitch_from_utau_mode1_track,
)
from .pitch_mode2 import (
    UtauMode2NotePitchData,
    UtauMode2TrackPitchData,
    pitch_from_utau_mode2_track,
)


@dataclasses.dataclass
class USTParser:
    options: InputOptions
    tempos: SortedKeyList = dataclasses.field(
        default_factory=functools.partial(SortedKeyList, key=operator.attrgetter("position"))
    )

    def parse_project(self, ust_project: UTAUProject) -> Project:
        if not len(ust_project.track):
            raise NoTrackError(_("UST project has no track"))
        tracks = []
        time_signatures = self.parse_time_signatures(ust_project.time_signatures)
        for ust_track in ust_project.track:
            track = SingingTrack()
            for (
                notes,
                mode1_track_pitch_data,
                mode2_track_pitch_data,
            ) in self.parse_notes(ust_project.tempo, ust_track.notes):
                track.note_list.extend(notes)
                if self.options.import_pitch:
                    synchronizer = TimeSynchronizer(list(self.tempos))
                    if ust_project.pitch_mode2:
                        track.edited_params.pitch.points.root.extend(
                            pitch_from_utau_mode2_track(
                                mode2_track_pitch_data, synchronizer, notes
                            ).points.root
                        )
                    else:
                        track.edited_params.pitch.points.root.extend(
                            pitch_from_utau_mode1_track(
                                mode1_track_pitch_data,
                                synchronizer,
                                notes,
                            ).points.root
                        )
            tracks.append(track)
        return Project(
            song_tempo_list=list(self.tempos),
            time_signature_list=time_signatures,
            track_list=tracks,
        )

    def parse_time_signatures(
        self, time_signatures: list[UTAUTimeSignature]
    ) -> list[TimeSignature]:
        if not len(time_signatures):
            return [TimeSignature()]
        return [
            TimeSignature(
                numerator=ts.numerator,
                denominator=ts.denominator,
                bar_index=ts.bar_index,
            )
            for ts in time_signatures
        ]

    def parse_notes(
        self, initial_tempo: float | None, notes: list[UTAUNote]
    ) -> Iterator[tuple[list[Note], UtauMode1TrackPitchData, UtauMode2TrackPitchData]]:
        note_list = []
        mode1_track_pitch_data = UtauMode1TrackPitchData()
        mode2_track_pitch_data = UtauMode2TrackPitchData()
        time = 0
        if initial_tempo is not None and not len(self.tempos):
            self.tempos.add(
                SongTempo(
                    position=time,
                    bpm=initial_tempo,
                )
            )
        if not len(self.tempos):
            # TODO: add warning
            self.tempos.add(
                SongTempo(
                    position=time,
                    bpm=DEFAULT_BPM,
                )
            )
        for note in notes:
            if note.tempo is not None:
                self.tempos.add(
                    SongTempo(
                        position=time,
                        bpm=note.tempo,
                    )
                )
            if note.lyric.upper() != "R":
                note_list.append(
                    Note(
                        lyric=note.lyric,
                        length=note.length,
                        start_pos=time,
                        key_number=note.note_num,
                    )
                )
                mode2_note_pitch_data = UtauMode2NotePitchData(
                    bpm=self.tempos[-1].bpm,
                    start=note.pbs[0] if len(note.pbs) and isinstance(note.pbs[0], float) else None,
                    start_shift=note.pbs[1]
                    if len(note.pbs) > 1 and isinstance(note.pbs[1], float)
                    else None,
                    widths=[x or 0 for x in (note.pbw or [])],
                    shifts=[x or 0 for x in (note.pby or [])],
                    curve_types=note.pbm or [],
                    vibrato_params=note.vbr,
                )
                mode2_track_pitch_data.notes.append(
                    mode2_note_pitch_data
                    if any(
                        getattr(mode2_note_pitch_data, key) is not None
                        for key in [
                            "start",
                            "start_shift",
                            "widths",
                            "shifts",
                            "curve_types",
                            "vibrato_params",
                        ]
                    )
                    else None
                )
                mode1_track_pitch_data.notes.append(
                    UtauMode1NotePitchData(note.pitch_bend_points)
                    if note.pitch_bend_points
                    else None
                )
            elif note_list:
                yield note_list, mode1_track_pitch_data, mode2_track_pitch_data
                note_list.clear()
                mode1_track_pitch_data.notes.clear()
                mode2_track_pitch_data.notes.clear()
            time += note.length
        if note_list:
            yield note_list, mode1_track_pitch_data, mode2_track_pitch_data



import pathlib
from typing import TYPE_CHECKING

from libresvip.extension import base as plugin_base
from libresvip.model.base import Project

from .options import OutputOptions
from .svg_generator import SvgGenerator

if TYPE_CHECKING:
    from drawsvg import Drawing


class SvgConverter(plugin_base.WriteOnlyConverterBase):
    def dump(self, path: pathlib.Path, project: Project, options: OutputOptions) -> None:
        drawing: Drawing = SvgGenerator(options).generate_project(project)
        path.write_bytes(drawing.as_svg().encode("utf-8"))



import contextlib
import hashlib
import pathlib
import sys
from typing import Any

try:
    __import__("Cryptodome")
except ImportError:
    with contextlib.suppress(ImportError):
        sys.modules["Cryptodome"] = __import__("Crypto")
from pydantic import Base64Bytes, Field, ValidationInfo, field_validator

from libresvip.core.compat import json, zstd
from libresvip.core.exceptions import UnsupportedProjectVersionError
from libresvip.model.base import BaseModel
from libresvip.utils.translation import gettext_lazy as _


class AcepDebug(BaseModel):
    os: str = "windows"
    platform: str = "pc"
    version: str = "10"


class AcepFile(BaseModel):
    compress_method: str = Field(default="zstd", alias="compressMethod")
    debug_info: AcepDebug = Field(
        default_factory=AcepDebug,
        alias="debugInfo",
    )
    salt: str = ""
    version: int = Field(default=1000)
    content: Base64Bytes

    @field_validator("version")
    @classmethod
    def version_validator(cls, value: int, _info: ValidationInfo) -> int:
        if value < 1000 and "Cryptodome" not in sys.modules:
            msg = _("Unsupported project version")
            raise UnsupportedProjectVersionError(msg)
        return value


def decrypt_acep_content_v1(content: bytes) -> bytes:
    from Cryptodome.Cipher import AES
    from Cryptodome.Util.Padding import unpad

    key_bytes = b"11956722077380335572"
    iv_bytes = b"1103392056537578664"
    key = hashlib.sha256(key_bytes).digest()
    iv = hashlib.md5(iv_bytes).digest()
    cipher = AES.new(key, AES.MODE_CBC, iv)
    return unpad(cipher.decrypt(content), AES.block_size, style="iso7816")


def decrypt_acep_content_v2(content: bytes, salt: str) -> bytes:
    from Cryptodome.Cipher import AES
    from Cryptodome.Util.Padding import unpad

    iv_bytes = str(int(salt[33:], 16) & 0x49A23B22DD28F042).encode()
    iv = hashlib.md5(iv_bytes).digest()
    key = b"\xb7\x4b\x57\x57\x5f\xb1\x4f\xc7\xec\xa9\x9c\x8b\x82\x53\x10\xfc\x4a\x33\x7b\x83\x7b\x12\x83\xe9\x0e\xe0\xef\x02\x20\x1c\x91\x12"
    cipher = AES.new(key, AES.MODE_CBC, iv)
    return unpad(cipher.decrypt(content), AES.block_size, style="iso7816")


def decompress_ace_studio_project(src: pathlib.Path) -> dict[str, Any]:
    acep_file = AcepFile.model_validate_json(src.read_bytes().decode("utf-8"))
    if acep_file.version == 1:
        content = decrypt_acep_content_v1(acep_file.content)
    elif acep_file.version == 2:
        content = decrypt_acep_content_v2(acep_file.content, acep_file.salt)
    else:
        content = acep_file.content
    decompressed = zstd.decompress(content)
    return json.loads(decompressed)


def compress_ace_studio_project(src: dict[str, Any], target: pathlib.Path) -> None:
    raw_content = json.dumps(src).encode()
    compressed = zstd.compress(raw_content)
    acep_file = AcepFile.model_construct(content=compressed)
    target.write_bytes(
        json.dumps(
            acep_file.model_dump(mode="json", by_alias=True),
            separators=(",", ":"),
        ).encode("utf-8")
    )



import dataclasses

from libresvip.core.tick_counter import shift_beat_list, shift_tempo_list
from libresvip.core.time_sync import TimeSynchronizer
from libresvip.model.base import (
    Note,
    ParamCurve,
    Points,
    Project,
    SingingTrack,
    SongTempo,
    TimeSignature,
)
from libresvip.model.pitch_simulator import PitchSimulator
from libresvip.model.point import Point
from libresvip.model.portamento import PortamentoPitch
from libresvip.model.relative_pitch_curve import RelativePitchCurve

from .model import (
    UFData,
    UFNotes,
    UFPitch,
    UFTempos,
    UFTimeSignatures,
    UFTracks,
)
from .options import InputOptions


@dataclasses.dataclass
class UFDataParser:
    options: InputOptions
    synchronizer: TimeSynchronizer = dataclasses.field(init=False)
    first_bar_length: int = dataclasses.field(init=False)

    def parse_project(self, ufdata_project: UFData) -> Project:
        uf_project = ufdata_project.project
        time_signature_list = self.parse_time_signatures(uf_project.time_signatures)
        self.first_bar_length = round(time_signature_list[0].bar_length())
        tick_prefix = int(time_signature_list[0].bar_length() * uf_project.measure_prefix)
        song_tempo_list = shift_tempo_list(self.parse_tempos(uf_project.tempos), tick_prefix)
        self.synchronizer = TimeSynchronizer(song_tempo_list)
        return Project(
            song_tempo_list=song_tempo_list,
            time_signature_list=shift_beat_list(
                time_signature_list,
                uf_project.measure_prefix,
            ),
            track_list=self.parse_tracks(uf_project.tracks, tick_prefix),
        )

    @staticmethod
    def parse_tempos(tempos: list[UFTempos]) -> list[SongTempo]:
        return [
            SongTempo(
                position=tempo.tick_position,
                bpm=tempo.bpm,
            )
            for tempo in tempos
        ]

    @staticmethod
    def parse_time_signatures(
        time_signatures: list[UFTimeSignatures],
    ) -> list[TimeSignature]:
        return [
            TimeSignature(
                bar_index=time_signature.measure_position,
                numerator=time_signature.numerator,
                denominator=time_signature.denominator,
            )
            for time_signature in time_signatures
        ]

    def parse_tracks(self, tracks: list[UFTracks], tick_prefix: int) -> list[SingingTrack]:
        track_list = []
        for track in tracks:
            singing_track = SingingTrack(
                title=track.name,
                note_list=self.parse_notes(track.notes, tick_prefix),
            )
            if self.options.import_pitch and track.pitch is not None:
                singing_track.edited_params.pitch = self.parse_pitch(
                    track.pitch, singing_track.note_list, tick_prefix
                )
            track_list.append(singing_track)
        return track_list

    def parse_pitch(self, pitch: UFPitch, note_list: list[Note], tick_prefix: int) -> ParamCurve:
        if pitch.is_absolute:
            pitch_points = [Point.start_point()]
            prev_point = pitch_points[-1]
            point = None
            for tick, value in zip(pitch.ticks, pitch.values):
                if value == 0 and prev_point.y == -100:
                    pitch_points.append(
                        Point(
                            x=tick + tick_prefix + self.first_bar_length,
                            y=-100,
                        )
                    )
                if value is not None:
                    point = Point(
                        x=tick + tick_prefix + self.first_bar_length,
                        y=round(value * 100),
                    )
                if point is not None:
                    pitch_points.append(point)
                if value == 0 and prev_point.y != -100:
                    pitch_points.append(
                        Point(
                            x=tick + tick_prefix + self.first_bar_length,
                            y=-100,
                        )
                    )
                if point is not None:
                    prev_point = point
            if prev_point.y == 0:
                pitch_points.append(prev_point._replace(y=-100))
            pitch_points.append(Point.end_point())
            return ParamCurve(points=Points(root=pitch_points))
        rel_pitch_points = [
            Point(
                x=tick + tick_prefix,
                y=round(value * 100),
            )
            for tick, value in zip(pitch.ticks, pitch.values)
            if value is not None
        ]
        pitch_simulator = PitchSimulator(
            synchronizer=self.synchronizer,
            portamento=PortamentoPitch.no_portamento(),
            note_list=note_list,
        )
        return RelativePitchCurve(self.first_bar_length).to_absolute(
            rel_pitch_points, pitch_simulator
        )

    @staticmethod
    def parse_notes(notes: list[UFNotes], tick_prefix: int) -> list[Note]:
        return [
            Note(
                start_pos=note.tick_on + tick_prefix,
                length=note.tick_off - note.tick_on,
                key_number=note.key,
                lyric=note.lyric,
            )
            for note in notes
        ]



import pathlib

from libresvip.core.compat import json
from libresvip.extension import base as plugin_base
from libresvip.model.base import Project

from .model import S5pProject
from .options import InputOptions, OutputOptions
from .synthv_editor_generator import SynthVEditorGenerator
from .synthv_editor_parser import SynthVEditorParser


class SynthVEditorConverter(plugin_base.SVSConverterBase):
    def load(self, path: pathlib.Path, options: InputOptions) -> Project:
        s5p_project = S5pProject.model_validate_json(
            path.read_bytes().decode("utf-8"), context={"path": path}
        )
        return SynthVEditorParser(options).parse_project(s5p_project)

    def dump(self, path: pathlib.Path, project: Project, options: OutputOptions) -> None:
        s5p_project = SynthVEditorGenerator(options).generate_project(project)
        path.write_bytes(
            json.dumps(
                s5p_project.model_dump(mode="json", by_alias=True, exclude_none=True),
                separators=(",", ":"),
            ).encode("utf-8"),
        )



from pydantic import BaseModel, ConfigDict
from xsdata_pydantic.fields import field


class Rootfile(BaseModel):
    """The rootfile element describes each top-level file in the MusicXML
    container.

    A MusicXML file used as a rootfile may have score-partwise, score-
    timewise, or opus as its document element. The required full-path
    attribute provides the path relative to the root folder of the zip
    file. The optional media-type attribute identifies the media type of
    different top-level root files. It is an error to have a non-
    MusicXML media-type value in the first rootfile in a rootfiles
    element. If no media-type value is present, a MusicXML file is
    assumed.
    """

    class Meta:
        name = "rootfile"

    model_config = ConfigDict(defer_build=True)
    full_path: str = field(
        metadata={
            "name": "full-path",
            "type": "Attribute",
            "required": True,
        }
    )
    media_type: str | None = field(
        default=None,
        metadata={
            "name": "media-type",
            "type": "Attribute",
        },
    )


class Rootfiles(BaseModel):
    """The rootfiles element includes the starting points for the different
    representations of a MusicXML score.

    The MusicXML root must be described in the first rootfile element.
    Additional rootfile elements can describe alternate versions such as
    PDF and audio files. When a MusicXML file contains both a score file
    and separate files for individual parts, the score file is the one
    that is referenced in the first rootfile element. The part files can
    be linked from within the score file and need not be listed here.
    """

    class Meta:
        name = "rootfiles"

    model_config = ConfigDict(defer_build=True)
    rootfile: list[Rootfile] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "min_occurs": 1,
        },
    )


class Container(BaseModel):
    """
    The container element with a container type is the root element of the META-
    INF/container.xml file.
    """

    class Meta:
        name = "container"

    model_config = ConfigDict(defer_build=True)
    rootfiles: Rootfiles = field(
        metadata={
            "type": "Element",
            "required": True,
        }
    )



"""
Implementation of the PEP 3156 Event-Loop with Qt.

Copyright (c) 2018 Gerard Marull-Paretas <gerard@teslabs.com>
Copyright (c) 2014 Mark Harviston <mark.harviston@gmail.com>
Copyright (c) 2014 Arve Knudsen <arve.knudsen@gmail.com>

BSD License
"""

# mypy: disable-error-code="attr-defined"

from __future__ import annotations

__author__ = (
    "Sam McCormack",
    "Gerard Marull-Paretas <gerard@teslabs.com>, "
    "Mark Harviston <mark.harviston@gmail.com>, "
    "Arve Knudsen <arve.knudsen@gmail.com>",
)
__all__ = ["QEventLoop", "QThreadExecutor"]

import asyncio
import itertools
import os
import sys
import time
from concurrent.futures import Future
from queue import Queue
from typing import TYPE_CHECKING, Any

from loguru import logger
from PySide6 import QtCore, QtWidgets
from typing_extensions import ParamSpec, Self, TypeVar, TypeVarTuple, Unpack

from ._common import _fileno, make_signaller

from __feature__ import snake_case, true_property  # isort: skip  # noqa: F401

if TYPE_CHECKING:
    from collections.abc import Awaitable, Callable
    from concurrent.futures import Executor
    from contextvars import Context
    from types import TracebackType

    from _typeshed import FileDescriptor

    _T = TypeVar("_T")
    _P = ParamSpec("_P")
    _Ts = TypeVarTuple("_Ts")

Slot = QtCore.Slot
QApplication = QtWidgets.QApplication


class _QThreadWorker(QtCore.QThread):
    """
    Read jobs from the queue and then execute them.

    For use by the QThreadExecutor
    """

    def __init__(
        self,
        queue: Queue[tuple[Future[_T], Callable[_P, _T], Any, Any] | None],
        num: int,
        stack_size: int | None = None,
    ) -> None:
        self.__queue = queue
        self.__num = num
        super().__init__()
        if stack_size is not None:
            self.stack_size = stack_size

    def run(self) -> None:
        queue = self.__queue
        while True:
            command = queue.get()
            if command is None:
                # Stopping...
                break

            future, callback, args, kwargs = command
            logger.debug(
                "#{} got callback {} with args {} and kwargs {} from queue",
                self.__num,
                callback,
                args,
                kwargs,
            )
            if future.set_running_or_notify_cancel():
                logger.debug("Invoking callback")
                try:
                    r = callback(*args, **kwargs)
                except Exception as err:
                    logger.debug("Setting Future exception: {}", err)
                    future.set_exception(err)
                else:
                    logger.debug("Setting Future result: {}", r)
                    future.set_result(r)
            else:
                logger.debug("Future was canceled")

        logger.debug("Thread #{} stopped", self.__num)

    def wait(self, *args: Any, **kwargs: Any) -> bool:
        logger.debug("Waiting for thread #{} to stop...", self.__num)
        return super().wait(*args, **kwargs)


class QThreadExecutor:
    """
    ThreadExecutor that produces QThreads.

    Same API as `concurrent.futures.Executor`

    >>> from qasync import QThreadExecutor
    >>> with QThreadExecutor(5) as executor:
    ...     f = executor.submit(lambda x: 2 + x, 2)
    ...     r = f.result()
    ...     assert r == 4
    """

    def __init__(self, max_workers: int = 10, stack_size: int | None = None) -> None:
        super().__init__()
        self.__max_workers = max_workers
        self.__queue: Queue[Any | None] = Queue()
        if stack_size is None:
            # Match cpython/Python/thread_pthread.h
            if sys.platform.startswith("darwin"):
                stack_size = 16 * 2**20
            elif sys.platform.startswith("freebsd"):
                stack_size = 4 * 2**20
            elif sys.platform.startswith("aix"):
                stack_size = 2 * 2**20
        self.__workers = [
            _QThreadWorker(self.__queue, i + 1, stack_size) for i in range(max_workers)
        ]
        self.__been_shutdown = False

        for w in self.__workers:
            w.start()

    def submit(self, callback: Callable[_P, _T], *args: _P.args, **kwargs: _P.kwargs) -> Future[_T]:
        if self.__been_shutdown:
            msg = "QThreadExecutor has been shutdown"
            raise RuntimeError(msg)
        future: Future[_T] = Future()
        logger.debug(
            "Submitting callback {} with args {} and kwargs {} to thread worker queue",
            callback,
            args,
            kwargs,
        )
        self.__queue.put((future, callback, args, kwargs))
        return future

    def shutdown(self, wait: bool = True, *, cancel_futures: bool = False) -> None:
        if self.__been_shutdown:
            msg = "QThreadExecutor has been shutdown"
            raise RuntimeError(msg)

        self.__been_shutdown = True

        logger.debug("Shutting down")
        for _ in range(len(self.__workers)):
            # Signal workers to stop
            self.__queue.put(None)
        if wait:
            for w in self.__workers:
                w.wait()

    def __enter__(self) -> Self:
        if self.__been_shutdown:
            msg = "QThreadExecutor has been shutdown"
            raise RuntimeError(msg)
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        self.shutdown()


class _SimpleTimer(QtCore.QObject):
    def __init__(self) -> None:
        super().__init__()
        self.__callbacks: dict[int, asyncio.Handle] = {}
        self._stopped = False
        self.__debug_enabled = False

    def add_callback(self, handle: asyncio.TimerHandle, delay: float = 0) -> asyncio.TimerHandle:
        timerid = self.start_timer(int(max(0, delay) * 1000))
        self.__log_debug("Registering timer id {}", timerid)
        assert timerid not in self.__callbacks
        self.__callbacks[timerid] = handle
        return handle

    def timer_event(self, event: QtCore.QTimerEvent) -> None:
        timerid = event.timer_id()
        self.__log_debug("Timer event on id {}", timerid)
        if self._stopped:
            self.__log_debug("Timer stopped, killing {}", timerid)
            self.kill_timer(timerid)
            del self.__callbacks[timerid]
        else:
            try:
                handle = self.__callbacks[timerid]
            except KeyError as e:
                self.__log_debug(e)
            else:
                if handle._cancelled:
                    self.__log_debug("Handle {} cancelled", handle)
                else:
                    self.__log_debug("Calling handle {}", handle)
                    handle._run()
            finally:
                del self.__callbacks[timerid]
            self.kill_timer(timerid)

    def stop(self) -> None:
        self.__log_debug("Stopping timers")
        self._stopped = True

    def set_debug(self, enabled: bool) -> None:
        self.__debug_enabled = enabled

    def __log_debug(self, *args: Any, **kwargs: Any) -> None:
        if self.__debug_enabled:
            logger.debug(*args, **kwargs)


class _QEventLoop(asyncio.AbstractEventLoop):
    """
    Implementation of asyncio event loop that uses the Qt Event loop.

    >>> import asyncio
    >>>
    >>> app = getfixture('application')
    >>>
    >>> async def xplusy(x, y):
    ...     await asyncio.sleep(.1)
    ...     assert x + y == 4
    ...     await asyncio.sleep(.1)
    >>>
    >>> loop = QEventLoop(app)
    >>> asyncio.set_event_loop(loop)
    >>> with loop:
    ...     loop.run_until_complete(xplusy(2, 2))

    If the event loop shall be used with an existing and already running QApplication
    it must be specified in the constructor via already_running=True
    In this case the user is responsible for loop cleanup with stop() and close()

    The set_running_loop parameter is there for backwards compatibility and does nothing.
    """

    def __init__(
        self,
        app: QtCore.QCoreApplication | None = None,
        set_running_loop: bool = False,
        already_running: bool = False,
    ) -> None:
        self.__app = app or QApplication.instance()
        assert self.__app is not None, "No QApplication has been instantiated"
        self.__is_running = False
        self.__debug_enabled = False
        self.__default_executor = None
        self.__exception_handler = None
        self._read_notifiers: dict[FileDescriptor, QtCore.QSocketNotifier] = {}
        self._write_notifiers: dict[FileDescriptor, QtCore.QSocketNotifier] = {}
        self._timer = _SimpleTimer()

        self.signaller = make_signaller(object, tuple)
        self.signaller.signal.connect(lambda callback, args: self.call_soon(callback, *args))

        assert self.__app is not None
        super().__init__()

        # We have to set __is_running to True after calling
        # super().__init__() because of a bug in BaseEventLoop.
        if already_running:
            self.__is_running = True

            # it must be ensured that all pre- and
            # postprocessing for the eventloop is done
            self._before_run_forever()
            self.__app.aboutToQuit.connect(self._after_run_forever)

            # for asyncio to recognize the already running loop
            asyncio.events._set_running_loop(self)

    def run_forever(self) -> None:
        """Run eventloop forever."""

        if self.__is_running:
            msg = "Event loop already running"
            raise RuntimeError(msg)

        self.__is_running = True
        self._before_run_forever()

        try:
            self.__log_debug("Starting Qt event loop")
            asyncio.events._set_running_loop(self)
            rslt = self.__app.exec()
            self.__log_debug("Qt event loop ended with result {}", rslt)
        finally:
            asyncio.events._set_running_loop(None)
            self._after_run_forever()
            self.__is_running = False

    def run_until_complete(self, future: Awaitable[_T]) -> _T:  # type: ignore[override]
        """Run until Future is complete."""

        if self.__is_running:
            msg = "Event loop already running"
            raise RuntimeError(msg)

        self.__log_debug("Running {} until complete", future)
        future = asyncio.ensure_future(future, loop=self)

        def stop(*args: Any) -> None:
            self.stop()

        future.add_done_callback(stop)
        try:
            self.run_forever()
        finally:
            future.remove_done_callback(stop)
        self.__app.process_events()  # run loop one last time to process all the events
        if not future.done():
            msg = "Event loop stopped before Future completed."
            raise RuntimeError(msg)

        self.__log_debug("Future {} finished running", future)
        return future.result()

    def stop(self) -> None:
        """Stop event loop."""
        if not self.__is_running:
            self.__log_debug("Already stopped")
            return

        self.__log_debug("Stopping event loop...")
        self.__is_running = False
        self.__app.exit()
        self.__log_debug("Stopped event loop")

    def is_running(self) -> bool:
        """Return True if the event loop is running, False otherwise."""
        return self.__is_running

    def close(self) -> None:
        """
        Release all resources used by the event loop.

        The loop cannot be restarted after it has been closed.
        """
        if self.is_running():
            msg = "Cannot close a running event loop"
            raise RuntimeError(msg)
        if self.is_closed():
            return

        self.__log_debug("Closing event loop...")
        if self.__default_executor is not None:
            self.__default_executor.shutdown()

        super().close()

        self._timer.stop()
        self.__app = None

        for notifier in itertools.chain(
            self._read_notifiers.values(), self._write_notifiers.values()
        ):
            notifier.set_enabled(False)

        self._read_notifiers.clear()
        self._write_notifiers.clear()

    def call_later(
        self,
        delay: float,
        callback: Callable[[Unpack[_Ts]], object],
        *args: Unpack[_Ts],
        context: Context | None = None,
    ) -> asyncio.TimerHandle:
        """Register callback to be invoked after a certain delay."""
        if asyncio.iscoroutinefunction(callback):
            msg = "coroutines cannot be used with call_later"
            raise TypeError(msg)
        if not callable(callback):
            msg = f"callback must be callable: {type(callback).__name__}"
            raise TypeError(msg)

        self.__log_debug(
            "Registering callback {} to be invoked with arguments {} after {} second(s)",
            callback,
            args,
            delay,
        )

        return self._add_callback(asyncio.TimerHandle(delay, callback, args, self, context=context))

    def _add_callback(self, handle: asyncio.TimerHandle, delay: float = 0) -> asyncio.TimerHandle:
        return self._timer.add_callback(handle, delay)

    def call_soon(
        self,
        callback: Callable[[Unpack[_Ts]], object],
        *args: Unpack[_Ts],
        context: Context | None = None,
    ) -> asyncio.Handle:
        """Register a callback to be run on the next iteration of the event loop."""
        return self.call_later(0, callback, *args, context=context)

    def call_at(
        self,
        when: float,
        callback: Callable[[Unpack[_Ts]], object],
        *args: Unpack[_Ts],
        context: Context | None = None,
    ) -> asyncio.TimerHandle:
        """Register callback to be invoked at a certain time."""
        return self.call_later(when - self.time(), callback, *args, context=context)

    def time(self) -> float:
        """Get time according to event loop's clock."""
        return time.monotonic()

    def _add_reader(
        self,
        fd: FileDescriptor,
        callback: Callable[[Unpack[_Ts]], object],
        *args: Unpack[_Ts],
    ) -> None:
        """Register a callback for when a file descriptor is ready for reading."""
        self._check_closed()

        try:
            existing = self._read_notifiers[fd]
        except KeyError:
            pass
        else:
            # this is necessary to avoid race condition-like issues
            existing.set_enabled(False)
            existing.activated["int"].disconnect()
            # will get overwritten by the assignment below anyways

        notifier = QtCore.QSocketNotifier(_fileno(fd), QtCore.QSocketNotifier.Type.Read)
        notifier.set_enabled(True)
        self.__log_debug("Adding reader callback for file descriptor {}", fd)
        notifier.activated["int"].connect(
            lambda: self.__on_notifier_ready(self._read_notifiers, notifier, fd, callback, *args)
        )
        self._read_notifiers[fd] = notifier

    def _remove_reader(self, fd: FileDescriptor) -> bool | None:
        """Remove reader callback."""
        if self.is_closed():
            return None

        self.__log_debug("Removing reader callback for file descriptor {}", fd)
        try:
            notifier = self._read_notifiers.pop(fd)
        except KeyError:
            return False
        else:
            notifier.set_enabled(False)
            return True

    def _add_writer(
        self,
        fd: FileDescriptor,
        callback: Callable[[Unpack[_Ts]], object],
        *args: Unpack[_Ts],
    ) -> None:
        """Register a callback for when a file descriptor is ready for writing."""
        self._check_closed()
        try:
            existing = self._write_notifiers[fd]
        except KeyError:
            pass
        else:
            # this is necessary to avoid race condition-like issues
            existing.set_enabled(False)
            existing.activated["int"].disconnect()
            # will get overwritten by the assignment below anyways

        notifier = QtCore.QSocketNotifier(
            _fileno(fd),
            QtCore.QSocketNotifier.Type.Write,
        )
        notifier.set_enabled(True)
        self.__log_debug("Adding writer callback for file descriptor {}", fd)
        notifier.activated["int"].connect(
            lambda: self.__on_notifier_ready(self._write_notifiers, notifier, fd, callback, *args)
        )
        self._write_notifiers[fd] = notifier

    def _remove_writer(self, fd: FileDescriptor) -> bool | None:
        """Remove writer callback."""
        if self.is_closed():
            return None

        self.__log_debug("Removing writer callback for file descriptor {}", fd)
        try:
            notifier = self._write_notifiers.pop(fd)
        except KeyError:
            return False
        else:
            notifier.set_enabled(False)
            return True

    def __notifier_cb_wrapper(
        self,
        notifiers: dict[FileDescriptor, QtCore.QSocketNotifier],
        notifier: QtCore.QSocketNotifier,
        fd: FileDescriptor,
        callback: Callable[[Unpack[_Ts]], _T],
        *args: Unpack[_Ts],
    ) -> None:
        # This wrapper gets called with a certain delay. We cannot know
        # for sure that the notifier is still the current notifier for
        # the fd.
        if notifiers.get(fd) is not notifier:
            return
        try:
            callback(*args)
        finally:
            # The notifier might have been overriden by the
            # callback. We must not re-enable it in that case.
            if notifiers.get(fd) is notifier:
                notifier.set_enabled(True)
            else:
                notifier.activated["int"].disconnect()

    def __on_notifier_ready(
        self,
        notifiers: dict[FileDescriptor, QtCore.QSocketNotifier],
        notifier: QtCore.QSocketNotifier,
        fd: FileDescriptor,
        callback: Callable[[Unpack[_Ts]], _T],
        *args: Unpack[_Ts],
    ) -> None:
        if fd not in notifiers:
            logger.warning(
                "Socket notifier for fd {} is ready, even though it should "
                "be disabled, not calling {} and disabling",
                fd,
                callback,
            )
            notifier.set_enabled(False)
            return

        # It can be necessary to disable QSocketNotifier when e.g. checking
        # ZeroMQ sockets for events
        assert notifier.is_enabled()
        self.__log_debug("Socket notifier for fd {} is ready", fd)
        notifier.set_enabled(False)
        self.call_soon(
            self.__notifier_cb_wrapper,
            notifiers,
            notifier,
            fd,
            callback,
            *args,
        )

    # Methods for interacting with threads.

    def call_soon_threadsafe(  # type: ignore[return-value]
        self,
        callback: Callable[[Unpack[_Ts]], _T],
        *args: Unpack[_Ts],
        context: Context | None = None,
    ) -> asyncio.Handle:
        """Thread-safe version of call_soon."""
        self.signaller.signal.emit(callback, args)

    def run_in_executor(
        self,
        executor: Executor,
        func: Callable[[Unpack[_Ts]], _T],
        *args: Unpack[_Ts],
    ) -> asyncio.Future[_T]:
        """Run callback in executor.

        If no executor is provided, the default executor will be used, which defers execution to
        a background thread.
        """
        self.__log_debug("Running callback {} with args {} in executor", func, args)
        if isinstance(func, asyncio.Handle):
            assert not args
            assert not isinstance(func, asyncio.TimerHandle)
            if func._cancelled:
                f: asyncio.Future[Any] = asyncio.Future()
                f.set_result(None)
                return f
            func, args = func.callback, func.args

        if executor is None:
            self.__log_debug("Using default executor")
            executor = self.__default_executor

        if executor is None:
            self.__log_debug("Creating default executor")
            executor = QThreadExecutor()
            self.__default_executor = executor

        return asyncio.wrap_future(executor.submit(func, *args))

    def set_default_executor(self, executor: Any) -> None:
        self.__default_executor = executor

    # Error handlers.

    def set_exception_handler(self, handler: asyncio.events._ExceptionHandler | None) -> None:
        self.__exception_handler = handler  # type: ignore[assignment]

    def default_exception_handler(self, context: asyncio.events._Context) -> None:
        """Handle exceptions.

        This is the default exception handler.

        This is called when an exception occurs and no exception
        handler is set, and can be called by a custom exception
        handler that wants to defer to the default behavior.

        context parameter has the same meaning as in
        `call_exception_handler()`.
        """
        self.__log_debug("Default exception handler executing")
        message = context.get("message") or "Unhandled exception in event loop"

        try:
            exception = context["exception"]
        except KeyError:
            exc_info = sys.exc_info()
        else:
            exc_info = (type(exception), exception, exception.__traceback__)

        log_lines = [message]
        log_lines.extend(
            f"{key}: {context[key]!r}"
            for key in [k for k in sorted(context) if k not in {"message", "exception"}]
        )
        self.__log_error("\n".join(log_lines), exc_info=exc_info)

    def call_exception_handler(self, context: asyncio.events._Context) -> None:
        if self.__exception_handler is None:
            try:
                self.default_exception_handler(context)
            except Exception:
                # Second protection layer for unexpected errors
                # in the default implementation, as well as for subclassed
                # event loops with overloaded "default_exception_handler".
                self.__log_error("Exception in default exception handler", exc_info=True)

            return

        try:
            self.__exception_handler(self, context)
        except Exception as exc:
            # Exception in the user set custom exception handler.
            try:
                # Let's try the default handler.
                self.default_exception_handler(
                    {
                        "message": "Unhandled error in custom exception handler",
                        "exception": exc,
                        "context": context,
                    }
                )
            except Exception:
                # Guard 'default_exception_handler' in case it's
                # overloaded.
                self.__log_error(
                    "Exception in default exception handler while handling an unexpected error "
                    "in custom exception handler",
                    exc_info=True,
                )

    # Debug flag management.

    def get_debug(self) -> bool:
        return self.__debug_enabled

    def set_debug(self, enabled: bool) -> None:
        super().set_debug(enabled)
        self.__debug_enabled = enabled
        self._timer.set_debug(enabled)

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        self.stop()
        self.close()

    def __log_debug(self, *args: Any, **kwargs: Any) -> None:
        if self.__debug_enabled:
            logger.debug(*args, **kwargs)

    @classmethod
    def __log_error(cls, *args: Any, **kwds: Any) -> None:
        # In some cases, the error method itself fails, don't have a lot of options in that case
        try:
            logger.exception(*args, **kwds)
        except Exception:
            sys.stderr.write(f"{args!r}, {kwds!r}\n")


if os.name == "nt":
    from ._windows import _ProactorEventLoop

    QIOCPEventLoop = type("QIOCPEventLoop", (_QEventLoop, _ProactorEventLoop), {})
    QEventLoop = QIOCPEventLoop
else:
    from ._unix import _SelectorEventLoop

    QSelectorEventLoop = type("QSelectorEventLoop", (_QEventLoop, _SelectorEventLoop), {})
    QEventLoop = QSelectorEventLoop



import pathlib

from libresvip.extension import base as plugin_base
from libresvip.model.base import Project

from .ass_generator import AssGenerator
from .options import OutputOptions


class AssConverter(plugin_base.WriteOnlyConverterBase):
    def dump(self, path: pathlib.Path, project: Project, options: OutputOptions) -> None:
        ssa_obj = AssGenerator(options).generate_project(project)
        path.write_bytes(ssa_obj.to_string("ass").encode(options.encoding))



from __future__ import annotations

from pydantic import Field

from libresvip.core.constants import TICKS_IN_BEAT
from libresvip.model.base import BaseModel
from libresvip.utils.text import uuid_str


class VoiceVoxSinger(BaseModel):
    engine_id: str = Field("074fc39e-678b-4c13-8916-ffca8d505d1d", alias="engineId")
    style_id: int = Field(3002, alias="styleId")


class VoiceVoxVoice(VoiceVoxSinger):
    speaker_id: str = Field("7ffcb7ce-00ec-4bdc-82cd-45a8889e43ff", alias="speakerId")
    style_id: int = Field(2, alias="styleId")


class VoiceVoxMora(BaseModel):
    text: str
    vowel: str
    vowel_length: float = Field(alias="vowelLength")
    pitch: float
    consonant: str | None = None
    consonant_length: float | None = Field(None, alias="consonantLength")


class VoiceVoxAccentPhrase(BaseModel):
    moras: list[VoiceVoxMora]
    accent: int
    mora: VoiceVoxMora | None = None
    is_interrogative: bool = Field(alias="isInterrogative")


class VoiceVoxQuery(BaseModel):
    accent_phrases: list[VoiceVoxAccentPhrase] = Field(alias="accentPhrases")
    speed_scale: float = Field(alias="speedScale")
    pitch_scale: float = Field(alias="pitchScale")
    intonation_scale: float = Field(alias="intonationScale")
    volume_scale: float = Field(alias="volumeScale")
    pre_phoneme_length: float = Field(alias="prePhonemeLength")
    post_phoneme_length: float = Field(alias="postPhonemeLength")
    pause_length: float | None = Field(None, alias="pauseLength")
    pause_length_scale: float = Field(1, alias="pauseLengthScale")
    output_sampling_rate: int = Field(alias="outputSamplingRate")
    output_stereo: bool = Field(alias="outputStereo")
    kana: str | None = None


class VoiceVoxAudioItem(BaseModel):
    text: str
    voice: VoiceVoxVoice
    query: VoiceVoxQuery
    preset_key: str = Field(default_factory=uuid_str, alias="presetKey")


class VoiceVoxTalk(BaseModel):
    audio_keys: list[str] = Field(default_factory=list, alias="audioKeys")
    audio_items: dict[str, VoiceVoxAudioItem] = Field(default_factory=dict, alias="audioItems")


class VoiceVoxTempo(BaseModel):
    position: int
    bpm: int


class VoiceVoxTimeSignature(BaseModel):
    measure_number: int = Field(alias="measureNumber")
    beats: int
    beat_type: int = Field(alias="beatType")


class VoiceVoxNote(BaseModel):
    id: str = Field(default_factory=uuid_str)
    position: int
    duration: int
    note_number: int = Field(alias="noteNumber")
    lyric: str


class VoiceVoxTrack(BaseModel):
    name: str
    singer: VoiceVoxSinger = Field(default_factory=VoiceVoxSinger)
    key_range_adjustment: int = Field(0, alias="keyRangeAdjustment")
    volume_range_adjustment: int = Field(0, alias="volumeRangeAdjustment")
    notes: list[VoiceVoxNote] = Field(default_factory=list)
    pitch_edit_data: list[float] = Field(alias="pitchEditData", default_factory=list)
    solo: bool = False
    mute: bool = False
    gain: float = 1.0
    pan: float = 0.0


class VoiceVoxSong(BaseModel):
    tpqn: int = TICKS_IN_BEAT
    tempos: list[VoiceVoxTempo] = Field(default_factory=list)
    time_signatures: list[VoiceVoxTimeSignature] = Field(
        alias="timeSignatures", default_factory=list
    )
    tracks: dict[str, VoiceVoxTrack] = Field(default_factory=dict)
    track_order: list[str] = Field(alias="trackOrder", default_factory=list)


class VoiceVoxProject(BaseModel):
    app_version: str = Field("0.21.1", alias="appVersion")
    talk: VoiceVoxTalk = Field(default_factory=VoiceVoxTalk)
    song: VoiceVoxSong = Field(default_factory=VoiceVoxSong)



import contextlib
import pathlib
import platform

from pydantic import ValidationInfo

from libresvip.core.warning_types import show_warning
from libresvip.utils.translation import gettext_lazy as _


def audio_path_validator(path: str, info: ValidationInfo) -> str:
    audio_path = pathlib.Path(path)
    if not audio_path.is_absolute() and info.context is not None:
        project_path: pathlib.Path | None
        if (project_path := info.context.get("path")) and not hasattr(project_path, "protocol"):
            audio_path = (project_path.parent / path).resolve()
            path = str(audio_path)
    return path


if platform.system() != "Emscripten":
    import xml.etree.ElementTree as ET

    from pymediainfo import MediaInfo
    from pymediainfo import Track as MediaInfoTrack

    def audio_track_info(
        file_path: str | pathlib.Path, only_wav: bool = False
    ) -> MediaInfoTrack | None:
        def filter_func(track: MediaInfoTrack) -> bool:
            return track.format == "PCM" if only_wav else (track.duration is not None)

        library_name = None
        if platform.system() == "Windows":
            python_compiler = platform.python_compiler()
            if python_compiler.startswith("GCC"):
                library_name = "libmediainfo-0.dll"

        if MediaInfo.can_parse(library_name):
            try:
                with contextlib.suppress(ET.ParseError, RuntimeError, ValueError):
                    media_info = MediaInfo.parse(file_path, library_file=library_name)
                    if not len(media_info.video_tracks):
                        return next(
                            (track for track in media_info.audio_tracks if filter_func(track)),
                            None,
                        )
            except FileNotFoundError:
                show_warning(_("Audio file not found: ") + f"{file_path}")
        return None
else:
    from xsdata.formats.dataclass.parsers.config import ParserConfig
    from xsdata_pydantic.bindings import XmlParser

    from .mediainfo_wasm import MediaInfo
    from .mediainfo_wasm import Track as MediaInfoTrack

    MEDIAINFO_JS_DIST_PATH = "https://fastly.jsdelivr.net/npm/mediainfo.js@latest/dist"

    try:
        import js
        from pyodide.ffi import run_sync

        if not hasattr(js, "MediaInfo"):
            js.importScripts(f"{MEDIAINFO_JS_DIST_PATH}/umd/index.min.js")

        def audio_track_info(
            file_path: str | pathlib.Path, only_wav: bool = False
        ) -> MediaInfoTrack | None:
            if isinstance(file_path, str):
                file_path = pathlib.Path(file_path)

            def filter_func(track: MediaInfoTrack) -> bool:
                return track.format == "PCM" if only_wav else (track.duration is not None)

            def read_chunk(size: int, offset: int) -> js.Uint8Array:
                chunk = content[offset : offset + size]
                js_buf = js.Uint8Array.new(len(chunk))
                js_buf.assign(chunk)
                return js_buf

            def locate_file(path: str, prefix: str) -> str:
                return f"{MEDIAINFO_JS_DIST_PATH}/MediaInfoModule.wasm"

            async def parse_media_info() -> str:
                media_info = await MediaInfo.mediaInfoFactory(format="XML", locateFile=locate_file)
                result = await media_info.analyzeData(len(content), read_chunk)
                media_info.close()
                return result

            if hasattr(js, "MediaInfo"):
                if file_path.exists():
                    with contextlib.suppress(RuntimeError, ValueError):
                        content = file_path.read_bytes()
                        xml_str = run_sync(parse_media_info())
                        media_info = XmlParser(
                            config=ParserConfig(fail_on_unknown_properties=False)
                        ).from_string(xml_str, MediaInfo)
                        if not len(media_info.video_tracks):
                            return next(
                                (track for track in media_info.audio_tracks if filter_func(track)),
                                None,
                            )
                else:
                    show_warning(_("Audio file not found: ") + f"{file_path}")
            return None
    except (ImportError, SystemError):

        def audio_track_info(
            file_path: str | pathlib.Path, only_wav: bool = False
        ) -> MediaInfoTrack | None:  # not implemented yet
            return None



from pydantic import BaseModel

from libresvip.model.option_mixins import (
    EnableInstrumentalTrackImportationMixin,
    EnablePitchImportationMixin,
)


class InputOptions(
    EnablePitchImportationMixin,
    EnableInstrumentalTrackImportationMixin,
    BaseModel,
):
    pass


class OutputOptions(BaseModel):
    pass



# mypy: disable-error-code="misc"
import enum
from typing import Annotated, Any

from pydantic import Field

from libresvip.model.base import BaseModel
from libresvip.utils.text import uuid_str


class PpsfLanguage(enum.IntEnum):
    JAPANESE: Annotated[int, Field(title="日本語")] = 0
    ENGLISH: Annotated[int, Field(title="English")] = 1
    SIMPLIFIED_CHINESE: Annotated[int, Field(title="简体中文")] = 4


class PpsfCurveType(enum.IntEnum):
    BORDER: Annotated[int, Field(title="Border")] = 0
    NORMAL: Annotated[int, Field(title="Normal")] = 1


class PpsfCurvePointSeq(BaseModel):
    border_type: int | None = Field(None, alias="border-type")
    note_index: int | None = Field(None, alias="note-index")
    region_index: int = Field(alias="region-index")
    edited_by_user: bool | None = Field(False, alias="edited-by-user")
    seg_array_id: int | None = Field(None, alias="seg-array-id")
    abs_value: int | None = Field(None, alias="abs-value")


class PpsfCurvePoint(BaseModel):
    plugin_descriptor: str | None = Field(None, alias="plugin-descriptor")
    sequence: list[PpsfCurvePointSeq] = Field(default_factory=list)
    sub_track_category: int = Field(alias="sub-track-category")
    sub_track_id: int = Field(alias="sub-track-id")


class PpsfSyllable(BaseModel):
    footer_text: str = Field("", alias="footer-text")
    header_text: str = Field("", alias="header-text")
    is_list_end: bool = Field(True, alias="is-list-end")
    is_list_top: bool = Field(True, alias="is-list-top")
    is_word_end: bool = Field(True, alias="is-word-end")
    is_word_top: bool = Field(True, alias="is-word-top")
    lyric_text: str = Field(alias="lyric-text")
    symbols_text: str = Field(alias="symbols-text")


class PpsfNote(BaseModel):
    language: PpsfLanguage = PpsfLanguage.JAPANESE
    region_index: int = Field(alias="region-index")
    syllables: list[PpsfSyllable] = Field(default_factory=list)
    event_index: int | None = None
    length: int | None = None
    muted: bool | None = False
    vibrato_preset_id: int | None = 0
    voice_color_id: int | None = 0
    voice_release_id: int | None = 0
    note_env_preset_id: int | None = 2
    note_gain_value: int | None = 0
    note_param_edited_stats: int | None = 0
    portamento_length: int | None = 180
    portamento_offset: int | None = -120
    vibrato_depth: int | None = 0
    vibrato_rate: int | None = 0
    auto_ai_consonant_length_enabled: bool | None = False
    auto_ai_f0_enabled: bool | None = True
    auto_ai_vibrato_blend_amount: int | None = None
    auto_ai_vibrato_blend_type: int | None = None
    auto_ai_vibrato_default_type: int | None = None
    auto_ai_vibrato_depth: int | None = 0
    auto_ai_vibrato_enabled: bool | None = False
    auto_ai_vibrato_offset: float | None = None
    auto_ai_vibrato_offset_by_user: bool | None = False


class PpsfRegion(BaseModel):
    auto_expand_left: bool | None = Field(None, alias="auto-expand-left")
    auto_expand_right: bool | None = Field(None, alias="auto-expand-right")
    length: int
    muted: bool = False
    name: str = ""
    position: int
    z_order: int = Field(0, alias="z-order")
    audio_event_index: int | None = Field(None, alias="audio-event-index")


class PpsfSubTrack(BaseModel):
    height: int
    plugin_descriptor: str | None = Field(None, alias="plugin-descriptor")
    sub_track_category: int = Field(alias="sub-track-category")
    sub_track_id: int = Field(alias="sub-track-id")


class PpsfParamPoint(BaseModel):
    curve_type: PpsfCurveType = PpsfCurveType.NORMAL
    pos: int
    value: int
    region_index: int | None = Field(None, alias="region-index")
    border_type: int | None = Field(None, alias="border-type")
    note_index: int | None = Field(None, alias="note-index")
    edited_by_user: bool | None = Field(None, alias="edited-by-user")
    seg_array_id: int | None = Field(None, alias="seg-array-id")


class PpsfBaseSequence(BaseModel):
    constant: int
    name: str
    sequence: list[PpsfParamPoint] = Field(default_factory=list)
    use_sequence: bool = False


class PpsfSeqParam(BaseModel):
    base_sequence: PpsfBaseSequence | None = Field(None, alias="base-sequence")
    layers: list[Any] | None = Field(default_factory=list)


class PpsfParameter(PpsfBaseSequence):
    default: int | None = None
    max_value: int | None = Field(None, alias="max")
    min_value: int | None = Field(None, alias="min")


class PpsfFsmEffect(BaseModel):
    parameters: list[PpsfParameter] = Field(default_factory=list)
    plugin_id: int = Field(alias="plugin-id")
    plugin_name: str = Field(alias="plugin-name")
    power_state: bool = Field(alias="power-state")
    program_name: str = Field(alias="program-name")
    program_number: int = Field(alias="program-number")
    version: str


class PpsfEventTrack(BaseModel):
    curve_points: list[PpsfCurvePoint] = Field(default_factory=list, alias="curve-points")
    fsm_effects: list[PpsfFsmEffect] | None = Field(default_factory=list, alias="fsm-effects")
    height: int = 64
    index: int
    mute_solo: int = Field(0, alias="mute-solo")
    notes: list[PpsfNote] = Field(default_factory=list)
    nt_envelope_preset_id: int | None = Field(None, alias="nt-envelope-preset-id")
    regions: list[PpsfRegion] = Field(default_factory=list)
    sub_tracks: list[PpsfSubTrack] = Field(default_factory=list, alias="sub-tracks")
    total_height: int = Field(64, alias="total-height")
    track_type: int = Field(alias="track-type")
    vertical_scale: float = Field(1, alias="vertical-scale")
    vertical_scroll: int = Field(0, alias="vertical-scroll")


class PpsfTempoTrack(BaseModel):
    height: int = 0
    vertical_scale: float = Field(2, alias="vertical-scale")
    vertical_scroll: int = Field(0, alias="vertical-scroll")


class PpsfTrackEditor(BaseModel):
    event_tracks: list[PpsfEventTrack] = Field(default_factory=list, alias="event-tracks")
    header_width: int = Field(264, alias="header-width")
    height: int = 720
    horizontal_scale: float = Field(0.08, alias="horizontal-scale")
    horizontal_scroll: int = Field(0, alias="horizontal-scroll")
    tempo_track: PpsfTempoTrack = Field(default_factory=PpsfTempoTrack, alias="tempo-track")
    user_markers: list[Any] | None = Field(default_factory=list, alias="user-markers")
    width: int = 1024
    x: int = 100
    y: int = 100


class PpsfLoopPoint(BaseModel):
    begin: int = 0
    enabled: bool | None = Field(False, validation_alias="enable")
    end: int = 7680


class PpsfMetronome(BaseModel):
    enabled: bool | None = Field(False, validation_alias="enable")
    wav: str = "null"


class PpsfGuiSettings(BaseModel):
    loop_point: PpsfLoopPoint | None = Field(default_factory=PpsfLoopPoint)
    metronome: PpsfMetronome | None = Field(default_factory=PpsfMetronome)
    ambient_enabled: bool | None = Field(None, alias="ambient-enabled")
    file_fullpath: str = Field("", alias="file-fullpath")
    playback_position: int = Field(0, alias="playback-position")
    project_length: int = Field(0, alias="project-length")
    track_editor: PpsfTrackEditor = Field(default_factory=PpsfTrackEditor, alias="track-editor")
    auto_connect_interval_msec: int | None = Field(None, alias="auto-connect-interval-msec")
    is_saved_by_nt2: bool | None = Field(None, alias="is-saved-by-nt2")


class PpsfFileAudioData(BaseModel):
    file_path: str = ""
    tempo: int = 0


class PpsfAudioTrackEvent(BaseModel):
    file_audio_data: PpsfFileAudioData
    playback_offset_sample: int = 0
    tick_length: int = 0
    tick_pos: int
    enabled: bool | None = Field(True, validation_alias="enable")


def default_gain() -> PpsfSeqParam:
    return PpsfSeqParam(
        base_sequence=PpsfBaseSequence(
            constant=750000,
            name="Gain",
        )
    )


def default_panpot() -> PpsfSeqParam:
    return PpsfSeqParam(
        base_sequence=PpsfBaseSequence(
            constant=0,
            name="Panpot",
        )
    )


class PpsfMixer(BaseModel):
    gain: PpsfSeqParam = Field(default_factory=default_gain)
    mixer_type: str = "stereo"
    panpot: PpsfSeqParam = Field(default_factory=default_panpot)


class PpsfAudioTrackItem(BaseModel):
    block_size: int = 279
    enabled: bool | None = Field(True, validation_alias="enable")
    events: list[PpsfAudioTrackEvent] = Field(default_factory=list)
    input_channel: int = 0
    mixer: PpsfMixer = Field(default_factory=PpsfMixer)
    name: str
    output_channel: int = 2
    sampling_rate: int = 48000


class PpsfBaseMeter(BaseModel):
    denomi: int = 4
    nume: int = 4


class PpsfMeter(PpsfBaseMeter):
    measure: int


class PpsfMeters(BaseModel):
    const: PpsfBaseMeter = Field(default_factory=PpsfBaseMeter)
    sequence: list[PpsfMeter] = Field(default_factory=list)
    use_sequence: bool = False


class PpsfSingerParam(BaseModel):
    breathiness: float = 0
    brightness: float = 0
    clearness: float = 0
    gender_factor: float = 0
    opening: float = 0


class PpsfSingerTableItem(BaseModel):
    cid1: str
    name: str
    param1: PpsfSingerParam = Field(default_factory=PpsfSingerParam)


class PpsfVocaloidTrackEvent(BaseModel):
    accent: int
    bend_depth: int
    bend_length: int
    decay: int
    fall_port: bool
    length: int
    lyric: str
    note_number: int
    opening: int
    pos: int
    protected: bool
    rise_port: bool
    symbols: str
    velocity: int
    vib_category: int
    vib_offset: int
    vib_depth: list[PpsfParamPoint] | None = None
    vib_rate: list[PpsfParamPoint] | None = None


class PpsfVocaloidTrackItem(BaseModel):
    events: list[PpsfVocaloidTrackEvent] = Field(default_factory=list)
    mixer: PpsfMixer = Field(default_factory=PpsfMixer)
    name: str
    parameters: list[PpsfParameter] = Field(default_factory=list)
    singer: int


class PpsfSinger(BaseModel):
    character_name: str = "miku"
    do_extrapolation: bool = False
    frame_shift: float = 0.001
    gender: int = 0
    language_id: PpsfLanguage = PpsfLanguage.JAPANESE
    library_id: str = Field(default_factory=uuid_str)
    name: str = ""
    singer_name: str = "HATSUNE MIKU NT Original"
    stationaly_type: str = "ParamedNDR"
    synthesis_version: str = "V20200915_H"
    tail_silent: float = 0.05
    vprm_morph_mode: str = "Linear"


class PpsfEnvelope(BaseModel):
    length: int = 0
    offset: int = 0
    points: list[PpsfParamPoint] = Field(default_factory=list)
    use_length: bool = True


def default_portamento_envelope() -> PpsfEnvelope:
    return PpsfEnvelope(length=0, offset=0)


class PpsfDvlTrackEvent(BaseModel):
    adjust_speed: bool | None = False
    attack_speed_rate: int = 800000
    consonant_rate: int = 950000
    consonant_speed_rate: int = 1000000
    enabled: bool | None = Field(True, validation_alias="enable")
    length: int
    lyric: str
    note_number: int
    note_off_pit_envelope: PpsfEnvelope | None = Field(default_factory=PpsfEnvelope)
    note_on_pit_envelope: PpsfEnvelope | None = Field(default_factory=PpsfEnvelope)
    portamento_envelope: PpsfEnvelope | None = Field(default_factory=default_portamento_envelope)
    vib_depth: PpsfEnvelope | None = None
    vib_rate: PpsfEnvelope | None = None
    vib_setting_id: int | None = None
    portamento_type: int = 5
    pos: int
    protected: bool = False
    release_speed_rate: int = 1400000
    symbols: str = ""
    vcl_like_note_off: bool | None = False
    is_consonant_length_by_user: bool | None = False

    @property
    def end_pos(self) -> int:
        return self.pos + self.length


class PpsfDvlTrackItem(BaseModel):
    enabled: bool | None = Field(True, validation_alias="enable")
    events: list[PpsfDvlTrackEvent] = Field(default_factory=list)
    mixer: PpsfMixer = Field(default_factory=PpsfMixer)
    name: str = "HATSUNE MIKU NT Original"
    parameters: list[PpsfSeqParam] = Field(default_factory=list)
    plugin_output_bus_index: int = 0
    singer: PpsfSinger = Field(default_factory=PpsfSinger)


class PpsfTempo(BaseModel):
    curve_type: PpsfCurveType = PpsfCurveType.NORMAL
    tick: int
    value: int


class PpsfTempos(BaseModel):
    const: int = 1200000
    sequence: list[PpsfTempo] = Field(default_factory=list)
    use_sequence: bool = False


class PpsfInnerProject(BaseModel):
    audio_track: list[PpsfAudioTrackItem] = Field(default_factory=list)
    block_size: int = 279
    loop_point: PpsfLoopPoint | None = None
    meter: PpsfMeters = Field(default_factory=PpsfMeters)
    metronome: PpsfMetronome | None = None
    name: str = ""
    sampling_rate: int = 48000
    singer_table: list[PpsfSingerTableItem] = Field(default_factory=list)
    tempo: PpsfTempos = Field(default_factory=PpsfTempos)
    vocaloid_track: list[PpsfVocaloidTrackItem] = Field(default_factory=list)
    dvl_track: list[PpsfDvlTrackItem] | None = None


class PpsfRoot(BaseModel):
    app_ver: str = "3.0.0.0"
    gui_settings: PpsfGuiSettings = Field(default_factory=PpsfGuiSettings)
    ppsf_ver: str = "3.2"
    project: PpsfInnerProject = Field(default_factory=PpsfInnerProject)


class PpsfProject(BaseModel):
    ppsf: PpsfRoot = Field(default_factory=PpsfRoot)



# mypy: disable-error-code="misc"
from enum import Enum
from typing import Annotated

from pydantic import Field

from libresvip.utils.translation import gettext_lazy as _


class PocketSingerLyricsLanguage(Enum):
    CHINESE: Annotated[
        str,
        Field(
            title=_("Chinese"),
        ),
    ] = "ch"
    JAPANESE: Annotated[
        str,
        Field(
            title=_("Japanese"),
        ),
    ] = "jp"
    ENGLISH: Annotated[
        str,
        Field(
            title=_("English"),
        ),
    ] = "en"



import dataclasses
import itertools
import operator
import re
from collections import defaultdict
from decimal import Decimal
from typing import cast

from libresvip.core.constants import TICKS_IN_BEAT
from libresvip.model.base import Project, SingingTrack, TimeSignature
from libresvip.utils.music_math import midi2note

from .model import KeyTick, MXmlMeasure, MXmlMeasureContent
from .models.enums import NoteTypeValue, StartStop, StartStopContinue, Step, Syllabic, TiedType
from .models.mxml4 import (
    Attributes,
    Direction,
    DirectionType,
    Encoding,
    Identification,
    Lyric,
    Metronome,
    Notations,
    PerMinute,
    Pitch,
    Rest,
    ScorePart,
    ScorePartwise,
    Slur,
    Sound,
    TextElementData,
    Tie,
    Tied,
    Time,
)
from .models.mxml4 import Note as MusicXMLNote
from .options import OutputOptions

DEFAULT_TICK_RATE_CEVIO = 2.0


@dataclasses.dataclass
class MusicXMLGenerator:
    options: OutputOptions

    def generate_project(self, project: Project) -> ScorePartwise:
        project_with_tick_rate_applied = self.apply_tick_rate(project)
        measures: dict[int, list[MXmlMeasure]] = defaultdict(list)
        for i, track in enumerate(project_with_tick_rate_applied.track_list):
            if isinstance(track, SingingTrack) and (
                key_ticks := self.get_key_ticks(i, track, project_with_tick_rate_applied)
            ):
                track_measures = self.get_measures(
                    key_ticks,
                    project_with_tick_rate_applied.time_signature_list,
                )
                measures[i].extend(track_measures)
        musicxml_project = ScorePartwise(
            identification=Identification(
                encoding=Encoding(
                    software=["LibreSVIP"],
                )
            )
        )
        for track_index, measures_part in measures.items():
            track_title = project_with_tick_rate_applied.track_list[track_index].title
            part_node, score_part_node = self.generate_part(measures_part, track_index, track_title)
            musicxml_project.part.append(part_node)
            musicxml_project.part_list.score_part.append(score_part_node)
        return musicxml_project

    def generate_part(
        self,
        measures: list[MXmlMeasure],
        track_index: int,
        track_title: str | None,
    ) -> tuple[ScorePartwise.Part, ScorePart]:
        part_id = f"P{track_index + 1}"
        part_node = ScorePartwise.Part(id=part_id)
        score_part = ScorePart(id=part_id)
        score_part.part_name.value = track_title or f"Track {track_index + 1}"
        prev_note_node = None
        prev_is_slur = False
        slur_number = 1
        for i, measure in enumerate(measures):
            slur_number = 1
            measure_node = ScorePartwise.Part.Measure(
                number=str(i + 1),
                attributes=[
                    Attributes(
                        divisions=Decimal.from_float(TICKS_IN_BEAT * DEFAULT_TICK_RATE_CEVIO)
                    )
                ],
            )
            if measure.time_signature is not None:
                measure_node.attributes.append(
                    Attributes(
                        time=[
                            Time(
                                beats=[str(measure.time_signature.numerator)],
                                beat_type=[str(measure.time_signature.denominator)],
                            )
                        ]
                    )
                )
            if measure.contents is not None:
                for content in measure.contents:
                    if content.bpm is not None:
                        sound_node, direction_node = self.generate_nodes_for_tempo(content)
                        measure_node.sound.append(sound_node)
                        measure_node.direction.append(direction_node)
                    elif content.note_type is not None and content.note is not None:
                        is_slur = content.note.lyric == "-"
                        if prev_note_node is not None:
                            if prev_note_node.notations:
                                prev_notation = prev_note_node.notations[0]
                            else:
                                prev_notation = Notations()
                                prev_note_node.notations.append(prev_notation)
                            if is_slur and not prev_is_slur:
                                prev_notation.slur.append(
                                    Slur(type_value=StartStopContinue.START, number=slur_number)
                                )
                            elif prev_is_slur:
                                prev_slur_number = next(
                                    (slur.number for slur in prev_notation.slur), slur_number
                                )
                                if is_slur:
                                    prev_notation.slur.append(
                                        Slur(
                                            type_value=StartStopContinue.CONTINUE,
                                            number=prev_slur_number,
                                        )
                                    )
                                else:
                                    prev_notation.slur.append(
                                        Slur(
                                            type_value=StartStopContinue.STOP,
                                            number=prev_slur_number,
                                        )
                                    )
                                    slur_number += 1
                        if note := self.generate_note_node(content):
                            note.voice = str(track_index + 2)
                            measure_node.note.append(note)
                            prev_note_node = note
                            prev_is_slur = is_slur
                    else:
                        if prev_note_node is not None and prev_is_slur:
                            if prev_note_node.notations:
                                prev_notation = prev_note_node.notations[0]
                            else:
                                prev_notation = Notations()
                                prev_note_node.notations.append(prev_notation)
                            prev_slur_number = next(
                                (slur.number for slur in prev_notation.slur), slur_number
                            )
                            prev_notation.slur.append(
                                Slur(type_value=StartStopContinue.STOP, number=prev_slur_number)
                            )
                            prev_note_node = None
                            prev_is_slur = False
                        measure_node.note.append(self.generate_rest_node(content))
            part_node.measure.append(measure_node)
        if prev_note_node is not None and prev_is_slur:
            if prev_note_node.notations:
                prev_notation = prev_note_node.notations[0]
            else:
                prev_notation = Notations()
                prev_note_node.notations.append(prev_notation)
            prev_slur_number = next((slur.number for slur in prev_notation.slur), slur_number)
            prev_notation.slur.append(
                Slur(type_value=StartStopContinue.STOP, number=prev_slur_number)
            )
        return part_node, score_part

    def generate_nodes_for_tempo(self, tempo: MXmlMeasureContent) -> tuple[Sound, Direction]:
        sound_node = Sound(tempo=Decimal.from_float(cast("float", tempo.bpm)))
        direction_node = Direction(
            direction_type=[
                DirectionType(
                    metronome=Metronome(
                        beat_unit=[NoteTypeValue.QUARTER],
                        per_minute=PerMinute(value=str(tempo.bpm)),
                    )
                )
            ]
        )
        return sound_node, direction_node

    def generate_rest_node(self, rest: MXmlMeasureContent) -> MusicXMLNote:
        return MusicXMLNote(rest=[Rest()], duration=[Decimal(rest.duration)])

    def generate_note_node(self, note: MXmlMeasureContent) -> MusicXMLNote | None:
        if note.note is None or note.note_type is None:
            return None
        note_node = MusicXMLNote(
            duration=[Decimal(note.duration)],
        )
        key_str = midi2note(note.note.key_number)
        if (octave_matcher := re.search(r"\d+$", key_str)) is not None:
            octave = int(octave_matcher.group())
        else:
            return None
        if (step_matcher := re.search(r"^[A-G]", key_str)) is not None:
            step = Step(step_matcher.group())
        else:
            return None
        alter = 1 if "#" in key_str else 0
        note_node.pitch.append(
            Pitch(
                step=step,
                alter=Decimal(alter) if alter != 0 else None,
                octave=octave,
            )
        )
        tie_type = {
            MXmlMeasureContent.NoteType.BEGIN: "start",
            MXmlMeasureContent.NoteType.END: "stop",
        }.get(note.note_type)
        if tie_type is not None:
            note_node.tie.append(Tie(type_value=StartStop(tie_type)))
            note_node.notations.append(Notations(tied=[Tied(type_value=TiedType(tie_type))]))
        note_node.lyric.append(
            Lyric(
                syllabic=[getattr(Syllabic, note.note_type.name)],
                text=[TextElementData(value=note.note.lyric)],
            )
        )
        return note_node

    @staticmethod
    def apply_tick_rate(project: Project) -> Project:
        return project.model_copy(
            update={
                "song_tempo_list": [
                    tempo.model_copy(
                        update={"position": int(tempo.position * DEFAULT_TICK_RATE_CEVIO)}
                    )
                    for tempo in project.song_tempo_list
                ],
                "track_list": [
                    track.model_copy(
                        update={
                            "note_list": [
                                note.model_copy(
                                    update={
                                        "start_pos": int(note.start_pos * DEFAULT_TICK_RATE_CEVIO),
                                        "length": int(note.length * DEFAULT_TICK_RATE_CEVIO),
                                    }
                                )
                                for note in track.note_list
                            ]
                        }
                    )
                    for track in project.track_list
                    if isinstance(track, SingingTrack)
                ],
            }
        )

    @staticmethod
    def get_key_ticks(track_index: int, track: SingingTrack, project: Project) -> list[KeyTick]:
        tempos = []
        if track_index == 0:
            tempos = [
                KeyTick(tick=tempo.position, tempo=tempo) for tempo in project.song_tempo_list
            ]
        note_starts = [KeyTick(tick=note.start_pos, note_start=note) for note in track.note_list]
        note_ends = [KeyTick(tick=note.end_pos, note_end=note) for note in track.note_list]
        return sorted(note_ends + tempos + note_starts, key=operator.attrgetter("tick"))

    @staticmethod
    def get_measures(
        key_ticks: list[KeyTick],
        time_signatures: list[TimeSignature],
    ) -> list[MXmlMeasure]:
        ticks_in_beat = round(TICKS_IN_BEAT * DEFAULT_TICK_RATE_CEVIO)
        measure_border_ticks = [0]
        measure = 0.0
        tick = 0.0
        prev_time_signature = TimeSignature()
        for time_signature in time_signatures:
            previous_measure = int(measure)
            ticks_in_measure = round(prev_time_signature.bar_length(ticks_in_beat))
            tick += ticks_in_measure * (time_signature.bar_index - measure)
            measure = time_signature.bar_index
            current_measure = int(measure)
            measure_border_ticks.extend(
                measure_border_ticks[-1] + ticks_in_measure
                for _ in range(current_measure - previous_measure)
            )
            prev_time_signature = time_signature
        last_tick = key_ticks[-1].tick
        if last_tick >= tick + (
            ticks_in_measure := round(prev_time_signature.bar_length(ticks_in_beat))
        ):
            previous_measure = int(measure)
            tick_diff = last_tick - tick
            measure += tick_diff / ticks_in_measure
            tick = last_tick
            current_measure = int(measure)
            measure_border_ticks.extend(
                measure_border_ticks[-1] + ticks_in_measure
                for _ in range(current_measure - previous_measure)
            )
        measure_border_ticks.append(
            measure_border_ticks[-1] + round(prev_time_signature.bar_length(ticks_in_beat))
        )

        key_ticks_with_measure_borders = [
            (
                border_pair,
                [
                    key_tick
                    for key_tick in key_ticks
                    if (
                        (border_pair[0] < key_tick.tick <= border_pair[1])
                        if key_tick.note_end is not None
                        else (border_pair[0] <= key_tick.tick < border_pair[1])
                    )
                ],
            )
            for border_pair in itertools.pairwise(measure_border_ticks)
        ]

        content_group_border_pair_map = {}
        ongoing_note_with_current_head = None
        for border_pair, key_tick_group in key_ticks_with_measure_borders:
            current_tick_in_measure = 0
            current_content_group: list[MXmlMeasureContent] = []
            for key_tick in key_tick_group:
                key_tick_relative = key_tick.tick - border_pair[0]
                if key_tick_relative > current_tick_in_measure:
                    if ongoing_note_with_current_head is None:
                        current_content_group.append(
                            MXmlMeasureContent.with_rest(
                                duration=key_tick_relative - current_tick_in_measure
                            )
                        )
                    current_tick_in_measure = key_tick_relative
                if key_tick.tempo is not None:
                    if ongoing_note_with_current_head is not None:
                        note, head = ongoing_note_with_current_head
                        current_content_group.append(
                            MXmlMeasureContent.with_note(
                                duration=key_tick.tick - head,
                                note=note,
                                note_type=MXmlMeasureContent.NoteType.BEGIN
                                if note.start_pos == head
                                else MXmlMeasureContent.NoteType.MIDDLE,
                            )
                        )
                        ongoing_note_with_current_head = note, key_tick.tick
                    current_content_group.append(
                        MXmlMeasureContent.with_tempo(bpm=key_tick.tempo.bpm)
                    )
                elif key_tick.note_start is not None:
                    ongoing_note_with_current_head = (
                        key_tick.note_start,
                        key_tick.tick,
                    )
                elif key_tick.note_end is not None and ongoing_note_with_current_head is not None:
                    note, head = ongoing_note_with_current_head
                    current_content_group.append(
                        MXmlMeasureContent.with_note(
                            duration=key_tick.note_end.end_pos - head,
                            note=key_tick.note_end,
                            note_type=MXmlMeasureContent.NoteType.SINGLE
                            if note.start_pos == head
                            else MXmlMeasureContent.NoteType.END,
                        )
                    )
                    ongoing_note_with_current_head = None
            rest_length = border_pair[1] - border_pair[0] - current_tick_in_measure
            if rest_length > 0:
                ongoing_note_at_measure_end = ongoing_note_with_current_head
                if ongoing_note_at_measure_end is None:
                    current_content_group.append(MXmlMeasureContent.with_rest(duration=rest_length))
                else:
                    note, head = ongoing_note_at_measure_end
                    current_content_group.append(
                        MXmlMeasureContent.with_note(
                            duration=border_pair[1] - head,
                            note=note,
                            note_type=MXmlMeasureContent.NoteType.BEGIN
                            if note.start_pos == head
                            else MXmlMeasureContent.NoteType.MIDDLE,
                        )
                    )
                    ongoing_note_with_current_head = note, border_pair[1]
            content_group_border_pair_map[border_pair] = current_content_group

        return [
            MXmlMeasure(
                tick_start=border_pair[0],
                length=border_pair[1] - border_pair[0],
                time_signature=next(
                    (ts for ts in time_signatures if ts.bar_index == index),
                    None,
                ),
                contents=contents,
            )
            for index, (border_pair, contents) in enumerate(
                sorted(
                    content_group_border_pair_map.items(),
                    key=lambda x: x[0][0],
                )
            )
        ]



from pydantic import BaseModel


class NotePositionParameters(BaseModel):
    point_1: tuple[float, float]
    point_2: tuple[float, float]
    inner_text: tuple[float, float]
    upper_text: tuple[float, float]
    lower_text: tuple[float, float]
    text_size: float



from pydantic import BaseModel, Field

from libresvip.utils.translation import gettext_lazy as _


class ProcessOptions(BaseModel):
    lyric_replacement_preset_name: str = Field(
        default="default",
        title=_("Name of lyric replacement preset"),
    )



import dataclasses
import itertools
import operator

from libresvip.core.constants import TICKS_IN_BEAT
from libresvip.core.time_sync import TimeSynchronizer
from libresvip.model.base import Note, ParamCurve, SongTempo
from libresvip.model.pitch_simulator import PitchSimulator
from libresvip.model.point import Point
from libresvip.model.portamento import PortamentoPitch
from libresvip.model.relative_pitch_curve import RelativePitchCurve
from libresvip.utils.search import find_last_index

from .constants import (
    MODE2_PITCH_MAX_POINT_COUNT,
    MODE2_PITCH_SAMPLING_INTERVAL_TICK,
)
from .interpolation import interpolate
from .model import UtauNoteVibrato, UTAUPitchBendMode
from .rdp_simplification import simplify_shape_to
from .vibrato_param import append_utau_note_vibrato


def milli_sec_from_tick(tick: int, bpm: float) -> float:
    return tick / TICKS_IN_BEAT * 60 / bpm * 1000


@dataclasses.dataclass
class UtauMode2NotePitchData:
    bpm: float
    start: float | None  # milliSec, None only if the note is not applied with pitch
    start_shift: float | None  # 10 cents
    widths: list[float] = dataclasses.field(default_factory=list)  # milliSec
    shifts: list[float] = dataclasses.field(default_factory=list)  # 10 cents
    curve_types: list[UTAUPitchBendMode] = dataclasses.field(default_factory=list)
    vibrato_params: UtauNoteVibrato | None = None


@dataclasses.dataclass
class UtauMode2TrackPitchData:
    notes: list[UtauMode2NotePitchData | None] = dataclasses.field(default_factory=list)


@dataclasses.dataclass
class NotePitchData:
    points: list[Point]
    offset: int
    bpm: float


def pitch_to_utau_mode2_track(
    pitch: ParamCurve, notes: list[Note], tempos: list[SongTempo]
) -> UtauMode2TrackPitchData:
    absolute_pitch = [
        point._replace(x=point.x - 1920) for point in pitch.points.root if point.y != -100
    ]

    def to_relative(from_: list[Point], key: int) -> list[Point]:
        return [Point(x=p.x, y=(p.y or key * 100) - key * 100) for p in from_]

    dot_pit_data = (
        [
            NotePitchData(
                to_relative(
                    [point for point in absolute_pitch if point.x < notes[0].end_pos],
                    notes[0].key_number,
                ),
                -min(
                    (point.x for point in absolute_pitch if point.x < 0),
                    default=0,
                ),
                bpm_for_note(tempos, notes[0]),
            ),  # first note
        ]
        + [
            NotePitchData(
                to_relative(
                    [point for point in absolute_pitch if note.start_pos <= point.x < note.end_pos],
                    note.key_number,
                ),
                (
                    next(
                        (point.x for point in absolute_pitch if point.x >= note.start_pos),
                        note.start_pos,
                    )
                    - note.start_pos
                ),
                bpm_for_note(tempos, note),
            )
            for note in notes[1:]
        ]
    )

    dot_pit_data_simplified = [
        NotePitchData(
            simplify_shape_to(x.points, MODE2_PITCH_MAX_POINT_COUNT),
            x.offset,
            x.bpm,
        )
        for x in dot_pit_data
    ]

    return UtauMode2TrackPitchData(
        notes=[
            UtauMode2NotePitchData(
                curr_note.bpm,
                milli_sec_from_tick(curr_note.offset, curr_note.bpm),
                curr_note.points[0].y / 10,
                [
                    milli_sec_from_tick(next_point.x - prev_point.x, curr_note.bpm)
                    for prev_point, next_point in zip(curr_note.points, curr_note.points[1:])
                ],
                [point.y / 10 for point in curr_note.points[1:]],
                [""] * (len(curr_note.points) - 1),
                None,
            )
            if curr_note.points
            else None
            for curr_note in dot_pit_data_simplified
        ]
    )


def pitch_from_utau_mode2_track(
    pitch_data: UtauMode2TrackPitchData,
    tick_time_transformer: TimeSynchronizer,
    notes: list[Note],
) -> ParamCurve:
    if pitch_data is None:
        return ParamCurve()
    pitch_points: list[Point] = []
    last_note: Note | None = None
    pending_pitch_points: list[Point] = []
    for note, note_pitch in zip(notes, pitch_data.notes):
        if note_pitch is not None:
            points = []
            note_start_in_millis = (
                tick_time_transformer.get_actual_secs_from_ticks(note.start_pos) * 1000
            )
            if note_pitch.start is not None:
                pos_in_millis = note_start_in_millis + note_pitch.start
                tick_pos = tick_time_transformer.get_actual_ticks_from_secs(pos_in_millis / 1000)
                start_shift = round(
                    (last_note.key_number - note.key_number) * 100
                    if last_note and note.start_pos == last_note.end_pos
                    else note_pitch.start_shift * 10
                    if note_pitch.start_shift is not None
                    else 0
                )
                points.append(Point(x=round(tick_pos), y=start_shift))
                for width, shift, curve_type in itertools.zip_longest(
                    note_pitch.widths,
                    note_pitch.shifts,
                    note_pitch.curve_types,
                ):
                    if width is None:
                        break
                    if shift is None:
                        shift = 0
                    if curve_type is None:
                        curve_type = ""
                    pos_in_millis += width
                    tick_pos = tick_time_transformer.get_actual_ticks_from_secs(
                        pos_in_millis / 1000
                    )
                    this_point = Point(x=round(tick_pos), y=round(shift * 10))
                    last_point = points[-1]
                    if this_point.x != last_point.x and this_point.y != last_point.y:
                        interpolated_point_list = interpolate(
                            last_point,
                            this_point,
                            curve_type,
                            MODE2_PITCH_SAMPLING_INTERVAL_TICK,
                        )
                        points.extend(interpolated_point_list[1:])
                    else:
                        points.append(this_point)
            pitch_points.extend(
                point
                for point in pending_pitch_points
                if point.x < (points[0].x if points else float("inf"))
            )
            pending_pitch_points = shape(
                append_utau_note_vibrato(
                    append_end_point(
                        append_start_point(
                            fix_points_at_last_note(
                                points,
                                note,
                                last_note,
                            ),
                            note,
                        ),
                        note,
                    ),
                    note_pitch.vibrato_params,
                    note,
                    tick_time_transformer,
                    MODE2_PITCH_SAMPLING_INTERVAL_TICK,
                )
            )
        last_note = note
    pitch_points.extend(pending_pitch_points)
    pitch_simulator = PitchSimulator(
        synchronizer=tick_time_transformer,
        portamento=PortamentoPitch.no_portamento(),
        note_list=notes,
    )
    return RelativePitchCurve(
        lower_bound=notes[0].start_pos,
        upper_bound=notes[-1].end_pos,
    ).to_absolute(pitch_points, pitch_simulator)


def fix_points_at_last_note(
    pitch_data: list[Point], this_note: Note, last_note: Note | None
) -> list[Point]:
    if last_note is None or last_note.end_pos != this_note.start_pos:
        return pitch_data
    fixed = [
        Point(
            x=point.x,
            y=point.y + (this_note.key_number - last_note.key_number) * 100,
        )
        if point.x < this_note.start_pos
        else point
        for point in pitch_data
    ]
    last_point = fixed[-1] if fixed else None
    return (
        [*fixed, Point(x=this_note.start_pos, y=0)]
        if last_point is not None and last_point.x < this_note.start_pos
        else fixed
    )


def append_start_point(pitch_data: list[Point], this_note: Note) -> list[Point]:
    first_point = pitch_data[0] if pitch_data else None
    if first_point is None:
        return [Point(x=this_note.start_pos, y=0)]
    elif first_point.x > this_note.start_pos:
        return [Point(x=this_note.start_pos, y=first_point.y), *pitch_data]
    else:
        return pitch_data


def append_end_point(pitch_data: list[Point], this_note: Note) -> list[Point]:
    last_point = pitch_data[-1] if pitch_data else None
    if last_point is None:
        return [Point(x=this_note.end_pos, y=0)]
    elif last_point.x < this_note.end_pos:
        return [*pitch_data, Point(this_note.end_pos, last_point.y)]
    else:
        return pitch_data


def shape(pitch_data: list[Point]) -> list[Point]:
    sorted_data = sorted(pitch_data, key=operator.attrgetter("x"))
    shaped_data: list[Point] = []
    for point in sorted_data:
        if shaped_data and shaped_data[-1].x == point.x:
            shaped_data[-1] = Point(x=shaped_data[-1].x, y=(shaped_data[-1].y + point.y) // 2)
        else:
            shaped_data.append(point)
    return shaped_data


def bpm_for_note(tempos: list[SongTempo], note: Note) -> float:
    tempo_index = find_last_index(tempos, lambda x: x.position <= note.start_pos)
    return tempos[tempo_index].bpm



import contextlib
import dataclasses

import mido_fix as mido

from libresvip.core.constants import TICKS_IN_BEAT
from libresvip.core.lyric_phoneme.chinese import get_pinyin_series
from libresvip.core.time_sync import TimeSynchronizer
from libresvip.core.warning_types import show_warning
from libresvip.model.base import (
    InstrumentalTrack,
    Note,
    ParamCurve,
    Project,
    SingingTrack,
    SongTempo,
    TimeSignature,
    Track,
)
from libresvip.utils.translation import gettext_lazy as _

from .constants import SUPPORTED_PINYIN
from .model import (
    GjgjBeatItems,
    GjgjBeatStyle,
    GjgjInstrumentalTrack,
    GjgjPoint,
    GjgjProject,
    GjgjSingingTrack,
    GjgjTempoMap,
    GjgjTempos,
    GjgjTimeSignature,
    GjgjTone,
    GjgjTrackVolume,
    GjgjVolumeMap,
)
from .options import OutputOptions
from .singers import DEFAULT_SINGER_ID, singer2id


@dataclasses.dataclass
class GjgjGenerator:
    options: OutputOptions
    max_note_id: int = dataclasses.field(init=False)
    first_bar_length: int = dataclasses.field(init=False)
    first_bar_bpm: float = dataclasses.field(init=False)
    param_sample_interval: int = dataclasses.field(init=False)
    time_synchronizer: TimeSynchronizer = dataclasses.field(init=False)

    def generate_project(self, project: Project) -> GjgjProject:
        self.max_note_id = 1
        self.time_synchronizer = TimeSynchronizer(project.song_tempo_list)
        gjgj_project = GjgjProject(
            gjgj_version=2,
            tempo_map=GjgjTempoMap(
                ticks_per_quarter_note=TICKS_IN_BEAT,
                tempos=self.generate_tempos(project.song_tempo_list),
                time_signature=self.generate_time_signatures(project.time_signature_list),
            ),
        )
        gjgj_project.tracks, gjgj_project.accompaniments = self.generate_tracks(project.track_list)
        return gjgj_project

    def generate_tempos(self, song_tempo_list: list[SongTempo]) -> list[GjgjTempos]:
        self.first_bar_bpm = song_tempo_list[0].bpm
        return [
            GjgjTempos(
                microseconds_per_quarter_note=mido.bpm2tempo(tempo.bpm),
                time=tempo.position,
            )
            for tempo in song_tempo_list
        ]

    def generate_time_signatures(
        self, time_signature_list: list[TimeSignature]
    ) -> list[GjgjTimeSignature]:
        self.first_bar_length = int(time_signature_list[0].bar_length())
        gjgj_time_signatures = []
        prev_ticks = 0
        for time_signature in time_signature_list:
            gjgj_time_signatures.append(
                GjgjTimeSignature(
                    numerator=time_signature.numerator,
                    denominator=time_signature.denominator,
                    time=prev_ticks,
                )
            )
            prev_ticks += round(time_signature.bar_index * time_signature.bar_length())
        return gjgj_time_signatures

    def generate_tracks(
        self, track_list: list[Track]
    ) -> tuple[list[GjgjSingingTrack], list[GjgjInstrumentalTrack]]:
        singing_tracks = []
        instrumental_tracks = []
        track_index = 1
        for track in track_list:
            if isinstance(track, SingingTrack):
                singing_tracks.append(self.generate_singing_track(track, track_index))
                track_index += 1
            elif isinstance(track, InstrumentalTrack):
                instrumental_tracks.append(self.generate_instrumental_track(track, track_index))
                track_index += 1
        return singing_tracks, instrumental_tracks

    def generate_singing_track(self, track: SingingTrack, track_index: int) -> GjgjSingingTrack:
        return GjgjSingingTrack(
            id_value=str(track_index),
            name=singer2id.get(self.options.singer or track.ai_singer_name, DEFAULT_SINGER_ID),
            master_volume=GjgjTrackVolume(mute=track.mute),
            beat_items=self.generate_notes(track.note_list),
            tone=self.generate_pitch(track.edited_params.pitch),
            volume_map=self.generate_volume(track.edited_params.volume),
        )

    def generate_instrumental_track(
        self, track: InstrumentalTrack, track_index: int
    ) -> GjgjInstrumentalTrack:
        return GjgjInstrumentalTrack(
            id_value=str(track_index),
            path=track.audio_file_path,
            offset=self.position_to_time(track.offset),
            master_volume=GjgjTrackVolume(mute=track.mute),
        )

    def position_to_time(self, position: int) -> int:
        if position > 0:
            return round(self.time_synchronizer.get_actual_secs_from_ticks(position) * 10000000)
        else:
            return round(position / TICKS_IN_BEAT * 60 / self.first_bar_bpm * 10000000)

    def generate_notes(self, note_list: list[Note]) -> list[GjgjBeatItems]:
        notes = []
        pinyin_list = get_pinyin_series([note.lyric for note in note_list])
        for note, pinyin in zip(note_list, pinyin_list):
            pronunciation = note.pronunciation or pinyin
            if pronunciation not in SUPPORTED_PINYIN:
                msg_prefx = _("Unsupported pinyin:")
                show_warning(f"{msg_prefx} {pronunciation}")
                pronunciation = ""
            if note.head_tag == "0":
                beat_style = GjgjBeatStyle.SIL
            elif note.head_tag == "V":
                beat_style = GjgjBeatStyle.SP
            else:
                beat_style = GjgjBeatStyle.NONE
            pre_time, post_time = 0.0, 0.0
            if note.edited_phones is not None:
                with contextlib.suppress(Exception):
                    if note.edited_phones.head_length_in_secs != -1:
                        note_start_pos_in_ticks = note.start_pos
                        note_start_pos_in_secs = self.time_synchronizer.get_actual_secs_from_ticks(
                            note_start_pos_in_ticks
                        )
                        phone_head_position_in_secs = (
                            note_start_pos_in_secs + note.edited_phones.head_length_in_secs
                        )
                        phone_head_position_in_ticks = (
                            self.time_synchronizer.get_actual_ticks_from_secs(
                                phone_head_position_in_secs
                            )
                        )
                        difference = note_start_pos_in_ticks - phone_head_position_in_ticks
                        pre_time = -difference * (2000 / 3) / TICKS_IN_BEAT

                    if note.edited_phones.mid_ratio_over_tail != -1:
                        post_time = (
                            -(note.length / (1 + note.edited_phones.mid_ratio_over_tail))
                            * (2000 / 3)
                            / TICKS_IN_BEAT
                        )
            notes.append(
                GjgjBeatItems(
                    id_value=self.max_note_id,
                    lyric=note.lyric,
                    pinyin=pronunciation,
                    start_tick=note.start_pos + self.first_bar_length,
                    duration=note.length,
                    track=note.key_number,
                    style=beat_style,
                    pre_time=pre_time,
                    post_time=post_time,
                )
            )
            self.max_note_id += 1
        return notes

    def generate_pitch(self, pitch: ParamCurve) -> GjgjTone:
        ticks_buffer = []
        value_buffer = []
        ori_prev_ticks = -100
        pitch_points: list[GjgjPoint] = []
        modify_ranges = []
        for point in pitch.points.root:
            ori_ticks, ori_value = point.x, point.y
            if ori_prev_ticks != ori_ticks:
                if ori_value != -100 and ori_ticks not in (
                    -192000,
                    1073741823,
                ):
                    ticks_buffer.append(ori_ticks - self.first_bar_length)
                    value_buffer.append(ori_value)
                elif len(ticks_buffer) > 0 and len(value_buffer) > 0:
                    pitch_points.extend(
                        GjgjPoint(
                            x=ticks_buffer[i],
                            y=value_buffer[i],
                        )
                        for i in range(len(ticks_buffer))
                    )
                    modify_ranges.append(
                        GjgjPoint(
                            x=ticks_buffer[0],
                            y=ticks_buffer[-1],
                        )
                    )
                    ticks_buffer.clear()
                    value_buffer.clear()
            ori_prev_ticks = ori_ticks
        return GjgjTone(
            modifies=pitch_points,
            modify_ranges=modify_ranges,
        )

    def generate_volume(self, volume_curve: ParamCurve) -> list[GjgjVolumeMap]:
        volume_curve = volume_curve.reduce_sample_rate(self.options.down_sample)
        volume_map = []
        ticks_buffer = []
        value_buffer = []
        prev_ticks = 0
        for volume in volume_curve.points.root[1:]:
            tick = volume.x
            if prev_ticks != tick:
                ori_value = volume.y
                value = (ori_value + 1000) / 1000
                if ori_value != 0:
                    ticks_buffer.append(tick - self.first_bar_length)
                    value_buffer.append(value)
                elif len(ticks_buffer) > 0 and len(value_buffer) > 0:
                    volume_map.append(
                        GjgjVolumeMap(
                            time=ticks_buffer[0] - 5,
                            volume=1,
                        )
                    )
                    volume_map.extend(
                        GjgjVolumeMap(
                            time=ticks_buffer[i],
                            volume=value_buffer[i],
                        )
                        for i in range(len(ticks_buffer))
                    )
                    volume_map.append(
                        GjgjVolumeMap(
                            time=ticks_buffer[-1] + 5,
                            volume=1,
                        )
                    )
                    ticks_buffer.clear()
                    value_buffer.clear()
        return volume_map



# mypy: disable-error-code="attr-defined"
import dataclasses
import operator

import more_itertools

from .model import PitchPoint, USTXProject, UVoicePart
from .time_axis import TimeAxis
from .utils import music_math


@dataclasses.dataclass
class BasePitchGenerator:
    project: dataclasses.InitVar[USTXProject]
    pitch_start: int = 0
    pitch_interval: int = 5
    time_axis: TimeAxis = dataclasses.field(init=False)

    def __post_init__(self, project: USTXProject) -> None:
        self.time_axis = TimeAxis()
        self.time_axis.build_segments(project)

    def base_pitch(self, part: UVoicePart) -> list[float]:
        pitches = [0.0] * (part.notes[-1].end // self.pitch_interval) if part.notes else []
        prev_note = None

        index = 0
        for note in part.notes:
            if (
                0 < index < len(pitches)
                and self.pitch_start + index * self.pitch_interval < note.position
            ):
                pitches[index] = pitches[index - 1]
                index += 1
            if (
                index < len(pitches)
                and (pad := (note.end - self.pitch_start) // self.pitch_interval - index) >= 0
            ):
                pitches[index : index + pad + 1] = [note.tone * 100] * (pad + 1)
                index += pad + 1
        index = max(1, index)
        if index < len(pitches):
            pitches[index:] = [pitches[index - 1]] * (len(pitches) - index)

        for note in part.notes:
            if note.vibrato.length <= 0 or note.vibrato.period is None:
                continue
            start_index = max(
                0,
                int((note.position - self.pitch_start) / self.pitch_interval),
            )
            end_index = min(
                len(pitches),
                (note.end - self.pitch_start) // self.pitch_interval,
            )
            n_period = note.vibrato.period / self.time_axis.ms_between_tick_pos(
                note.position, note.end
            )
            for i in range(start_index, end_index):
                n_pos = (self.pitch_start + i * self.pitch_interval - note.position) / note.duration
                point = note.vibrato.evaluate(n_pos, n_period, note)
                pitches[i] = point[1] * 100

        for note in part.notes:
            pitch_points = list(
                more_itertools.unique_justseen(
                    (
                        PitchPoint(
                            x=self.time_axis.ms_pos_to_tick_pos(
                                self.time_axis.tick_pos_to_ms_pos(part.position + note.position)
                                + point.x
                            )
                            - part.position,
                            y=point.y * 10 + note.tone * 100,
                            shape=point.shape,
                        )
                        for point in note.pitch.data
                    ),
                    key=operator.attrgetter("x"),
                )
            )
            if not pitch_points:
                pitch_points.append(PitchPoint(x=note.position, y=note.tone * 100))
                pitch_points.append(PitchPoint(x=note.end, y=note.tone * 100))
            if note == part.notes[0] and pitch_points[0].x > self.pitch_start:
                pitch_points.insert(0, PitchPoint(x=self.pitch_start, y=pitch_points[0].y))
            elif pitch_points[0].x > note.position:
                pitch_points.insert(0, PitchPoint(x=note.position, y=pitch_points[0].y))
            if pitch_points[-1].x < note.end:
                pitch_points.append(PitchPoint(x=note.end, y=pitch_points[-1].y))
            prev_point = pitch_points[0]
            index = max(0, int((prev_point.x - self.pitch_start) / self.pitch_interval))
            for point in pitch_points[1:]:
                x = self.pitch_start + index * self.pitch_interval
                while x <= point.x and index < len(pitches):
                    pitch = music_math.interpolate_shape(
                        (prev_point.x, prev_point.y),
                        (point.x, point.y),
                        x,
                        prev_point.shape or "lo",
                    )
                    base_pitch = (
                        prev_note.tone * 100
                        if prev_note is not None and x <= prev_note.end
                        else note.tone * 100
                    )
                    pitches[index] += pitch - base_pitch
                    index += 1
                    x += self.pitch_interval
                prev_point = point
            prev_note = note

        return pitches



import asyncio
from collections.abc import Callable
from typing import Any

from nicegui.elements.mixins.color_elements import BackgroundColorElement
from nicegui.elements.mixins.disableable_element import DisableableElement
from nicegui.elements.mixins.text_element import TextElement
from nicegui.events import ClickEventArguments, handle_event


class QFab(TextElement, DisableableElement, BackgroundColorElement):
    def __init__(
        self,
        text: str = "",
        *,
        on_click: Callable[..., Any] | None = None,
        color: str | None = "primary",
        icon: str | None = None,
    ) -> None:
        super().__init__(tag="q-fab", text=text, background_color=color)

        if icon:
            self._props["icon"] = icon

        if on_click:
            self.on(
                "click",
                lambda _: handle_event(
                    on_click,
                    ClickEventArguments(sender=self, client=self.client),
                ),
            )

    def _text_to_model_text(self, text: str) -> None:
        self._props["label"] = text

    async def clicked(self) -> None:
        """Wait until the button is clicked."""
        event = asyncio.Event()
        self.on("click", event.set)
        await self.client.connected()
        await event.wait()


class QFabAction(TextElement, DisableableElement, BackgroundColorElement):
    def __init__(
        self,
        text: str = "",
        *,
        on_click: Callable[..., Any] | None = None,
        color: str | None = "primary",
        icon: str | None = None,
    ) -> None:
        super().__init__(tag="q-fab-action", text=text, background_color=color)

        if icon:
            self._props["icon"] = icon

        if on_click:
            self.on(
                "click",
                lambda _: handle_event(
                    on_click,
                    ClickEventArguments(sender=self, client=self.client),
                ),
            )

    def _text_to_model_text(self, text: str) -> None:
        self._props["label"] = text

    async def clicked(self) -> None:
        """Wait until the button is clicked."""
        event = asyncio.Event()
        self.on("click", event.set)
        await self.client.connected()
        await event.wait()
